from __future__ import absolute_import, division, print_function
import numpy as np
import time
import copy
from scipy.optimize import fsolve, minimize, basinhopping, root
from scipy import optimize, integrate, interpolate
import scipy.constants
from operator import itemgetter
try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

from NuRadioMC.utilities import units
from NuRadioMC.utilities import attenuation as attenuation_util

import logging
logging.basicConfig()

# check if CPP implementation is available
cpp_available = False

try:
    from NuRadioMC.SignalProp.CPPAnalyticRayTracing import wrapper
    cpp_available = True
    print("using CPP version of ray tracer")
except:
    print("trying to compile the CPP extension on-the-fly")
    try:
        import subprocess
        import os
        subprocess.call(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 "install.sh"))
        from NuRadioMC.SignalProp.CPPAnalyticRayTracing import wrapper
        cpp_available = True
        print("compilation was sucessful, using CPP version of ray tracer")
    except:
        print("compilation was not sucessful, using python version of ray tracer")
        cpp_available = False

"""
analytic ray tracing solution
"""
speed_of_light = scipy.constants.c * units.m / units.s

solution_types = {1: 'direct',
                  2: 'refracted',
                  3: 'reflected'}


@lru_cache(maxsize=32)
def get_z_deep(ice_params):
    """
    Calculates the z_deep needed for integral along the homogeneous ice
    to know the path length or the times. We obtain the depth for which
    the index of refraction is 0.035% away of that of deep ice. This
    calculation assumes a monotonically increasing index of refraction
    with negative depth.
    """
    n_ice, z_0, delta_n = ice_params

    def diff_n_ice(z):

        rel_diff = 2e-5
        return delta_n * np.exp(z / z_0) / n_ice - rel_diff

    res =  optimize.root(diff_n_ice, -100 * units.m).x[0]
    print("zdeep = {:.2f}".format(res))
    return res


class ray_tracing_2D():

    def __init__(self, medium, attenuation_model="SP1",
                 log_level=logging.WARNING,
                 n_frequencies_integration=25,
                 use_optimized_start_values=False):
        """
        initialize 2D analytic ray tracing class

        Parameters
        ----------
        medium: NuRadioMC.utilities.medium class
            details of the medium
        attenuation_model: string
            specifies which attenuation model to use (default 'SP1')
        log_level: logging.loglevel object
            controls verbosity (default WARNING)
        n_frequencies_integration: int
            specifies for how many frequencies the signal attenuation is being calculated
        use_optimized_start_value: bool
            if True, the initial C_0 paramter (launch angle) is set to the ray that skims the surface
            (default: False)

        """
        self.medium = medium
        if(not hasattr(self.medium, "reflection")):
            self.medium.reflection = None
        
        self.attenuation_model = attenuation_model
        if(not self.attenuation_model in attenuation_util.model_to_int):
            raise NotImplementedError("attenuation model {} is not implemented".format(self.attenuation_model))
        self.attenuation_model_int = attenuation_util.model_to_int[self.attenuation_model]
        self.__b = 2 * self.medium.n_ice
        self.__logger = logging.getLogger('ray_tracing_2D')
        self.__logger.setLevel(log_level)
        self.__n_frequencies_integration = n_frequencies_integration
        self.__use_optimized_start_values = use_optimized_start_values

    def n(self, z):
        """
        refractive index as a function of depth
        """
        res = self.medium.n_ice - self.medium.delta_n * np.exp(z / self.medium.z_0)
    #     if(type(z) is float):
    #         if(z > 0):
    #             return 1.
    #         else:
    #             return res
    #     else:
    #         res[z > 0] = 1.
        return res

    def get_gamma(self, z):
        """
        transforms z coordinate into gamma
        """
        return self.medium.delta_n * np.exp(z / self.medium.z_0)

    def get_turning_point(self, c):
        """
        calculate the turning point, i.e. the maximum of the ray tracing path;
        parameter is c = self.medium.n_ice ** 2 - C_0 ** -2
        
        This is either the point of reflection off the ice surface
        or the point where the saddle point of the ray (transition from upward to downward going)
        
        Technically, the turning point is set to z=0 if the saddle point is above the surface. 
        
        Parameters
        ----------
        c: float
            related to C_0 parameter via c = self.medium.n_ice ** 2 - C_0 ** -2
            
        Returns
        ----------
        typle (gamma, z coordinate of turning point)
        """
        gamma2 = self.__b * 0.5 - (0.25 * self.__b ** 2 - c) ** 0.5  # first solution discarded
        z2 = np.log(gamma2 / self.medium.delta_n) * self.medium.z_0
        
        if(z2 > 0):
            z2 = 0  # a reflection is just a turning point at z = 0, i.e. cases 2) and 3) are the same
            gamma2 = self.get_gamma(z2)

        return gamma2, z2

    def get_y_turn(self, C_0, x1):
        """
        calculates the y-coordinate of the turning point. This is either the point of reflection off the ice surface
        or the point where the saddle point of the ray (transition from upward to downward going)
        
        Parameters
        ----------
        C_0: float
            C_0 parameter of function
        x1: typle
            (y, z) start position of ray
        """
        c = self.medium.n_ice ** 2 - C_0 ** -2
        gamma_turn, z_turn = self.get_turning_point(c)
        C_1 = x1[0] - self.get_y_with_z_mirror(x1[1], C_0)
        y_turn = self.get_y(gamma_turn, C_0, C_1)
        return y_turn
    
    def get_C_1(self, x1, C_0):
        """
        calculates constant C_1 for a given C_0 and start point x1
        """
        return x1[0] - self.get_y_with_z_mirror(x1[1], C_0)

    def get_c(self, C_0):
        return self.medium.n_ice ** 2 - C_0 ** -2

    def get_C0_from_log(self, logC0):
        """
        transforms the fit parameter C_0 so that the likelihood looks better
        """
        return np.exp(logC0) + 1. / self.medium.n_ice

    def get_y(self, gamma, C_0, C_1):
        """
        analytic form of the ray tracing part given an exponential index of refraction profile

        Parameters
        -------
        gamma: (float or array)
            gamma is a function of the depth z
        C_0: (float)
            first parameter
        C_1: (float)
            second parameter
        """
        c = self.medium.n_ice ** 2 - C_0 ** -2
        # we take the absolute number here but we only evaluate the equation for
        # positive outcome. This is to prevent rounding errors making the root
        # negative
        root = np.abs(gamma ** 2 - gamma * self.__b + c)
        logargument = gamma / (2 * c ** 0.5 * (root) ** 0.5 - self.__b * gamma + 2 * c)
        if(np.sum(logargument <= 0)):
            self.__logger.debug('log = {}'.format(logargument))
        result = self.medium.z_0 * (self.medium.n_ice ** 2 * C_0 ** 2 - 1) ** -0.5 * np.log(logargument) + C_1
        return result

    def get_y_diff(self, z_raw, C_0):
        """
        derivative dy(z)/dz
        """
        z = self.get_z_unmirrored(z_raw, C_0)
        c = self.medium.n_ice ** 2 - C_0 ** -2
        B = (0.2e1 * np.sqrt(c) * np.sqrt(-self.__b * self.medium.delta_n * np.exp(z / self.medium.z_0) + self.medium.delta_n ** 
                                          2 * np.exp(0.2e1 * z / self.medium.z_0) + c) - self.__b * self.medium.delta_n * np.exp(z / self.medium.z_0) + 0.2e1 * c)
        D = self.medium.n_ice ** 2 * C_0 ** 2 - 1
        E1 = -self.__b * self.medium.delta_n * np.exp(z / self.medium.z_0)
        E2 = self.medium.delta_n ** 2 * np.exp(0.2e1 * z / self.medium.z_0)
        E = (E1 + E2 + c)
        res = (-np.sqrt(c) * np.exp(z / self.medium.z_0) * self.__b * self.medium.delta_n + 0.2e1 * np.sqrt(-self.__b * self.medium.delta_n * np.exp(z / 
                                                                                                                                                     self.medium.z_0) + self.medium.delta_n ** 2 * np.exp(0.2e1 * z / self.medium.z_0) + c) * c + 0.2e1 * c ** 1.5) / B * E ** -0.5 * (D ** (-0.5))

        if(z != z_raw):
            res *= -1
        return res

    def get_y_with_z_mirror(self, z, C_0, C_1=0):
        """
        analytic form of the ray tracing part given an exponential index of refraction profile

        this function automatically mirrors z values that are above the turning point,
        so that this function is defined for all z

        Parameters
        -------
        z: (float or array)
            depth z
        C_0: (float)
            first parameter
        C_1: (float)
            second parameter
        """
        c = self.medium.n_ice ** 2 - C_0 ** -2
        gamma_turn, z_turn = self.get_turning_point(c)
        y_turn = self.get_y(gamma_turn, C_0, C_1)
        if(not hasattr(z, '__len__')):
            if(z < z_turn):
                gamma = self.get_gamma(z)
                return self.get_y(gamma, C_0, C_1)
            else:
                gamma = self.get_gamma(2 * z_turn - z)
                return 2 * y_turn - self.get_y(gamma, C_0, C_1)
        else:
            mask = z < z_turn
            res = np.zeros_like(z)
            zs = np.zeros_like(z)
            gamma = self.get_gamma(z[mask])
            zs[mask] = z[mask]
            res[mask] = self.get_y(gamma, C_0, C_1)
            gamma = self.get_gamma(2 * z_turn - z[~mask])
            res[~mask] = 2 * y_turn - self.get_y(gamma, C_0, C_1)
            zs[~mask] = 2 * z_turn - z[~mask]

            self.__logger.debug('turning points for C_0 = {:.2f}, b= {:.2f}, gamma = {:.4f}, z = {:.1f}, y_turn = {:.0f}'.format(
                C_0, self.__b, gamma_turn, z_turn, y_turn))
            return res, zs

    def get_z_mirrored(self, x1, x2, C_0):
        """
        calculates the mirrored x2 position so that y(z) can be used as a continuous function
        """
        c = self.medium.n_ice ** 2 - C_0 ** -2
        C_1 = x1[0] - self.get_y_with_z_mirror(x1[1], C_0)
        gamma_turn, z_turn = self.get_turning_point(c)
        y_turn = self.get_y(gamma_turn, C_0, C_1)
        zstart = x1[1]
        zstop = x2[1]
        if(y_turn < x2[0]):
            zstop = zstart + np.abs(z_turn - x1[1]) + np.abs(z_turn - x2[1])
        x2_mirrored = [x2[0], zstop]
        return x2_mirrored

    def get_z_unmirrored(self, z, C_0):
        """
        calculates the unmirrored z position
        """
        c = self.medium.n_ice ** 2 - C_0 ** -2
        gamma_turn, z_turn = self.get_turning_point(c)
        z_unmirrored = z
        if(z > z_turn):
            z_unmirrored = 2 * z_turn - z
        return z_unmirrored

    def ds(self, t, C_0):
        """
        helper to calculate line integral
        """
        return (self.get_y_diff(t, C_0) ** 2 + 1) ** 0.5

    def get_path_length(self, x1, x2, C_0, reflection=0, reflection_case=1):
        tmp = 0
        for iS, segment in enumerate(self.get_path_segments(x1, x2, C_0, reflection, reflection_case)):
            if(iS == 0 and reflection_case == 2):  # we can only integrate upward going rays, so if the ray starts downwardgoing, we need to mirror
                x11, x1, x22, x2, C_0, C_1 = segment
                x1t = copy.copy(x11)
                x2t = copy.copy(x2)
                x1t[1] = x2[1]
                x2t[1] = x11[1]
                x2 = x2t
                x1 = x1t
            else:
                x11, x1, x22, x2, C_0, C_1 = segment
            x2_mirrored = self.get_z_mirrored(x1, x2, C_0)
            gamma_turn, z_turn = self.get_turning_point(self.medium.n_ice ** 2 - C_0 ** -2)
            points = None
            if(x1[1] < z_turn and z_turn < x2_mirrored[1]):
                points = [z_turn]
            path_length = integrate.quad(self.ds, x1[1], x2_mirrored[1], args=(C_0), points=points, epsabs=1e-4, epsrel=1.49e-08, limit=50)
            self.__logger.info("calculating path length ({}) from ({:.0f}, {:.0f}) to ({:.2f}, {:.2f}) = ({:.2f}, {:.2f}) = {:.2f} m".format(solution_types[self.determine_solution_type(x1, x2, C_0)], x1[0], x1[1], x2[0], x2[1],
                                                                                                                                        x2_mirrored[0],
                                                                                                                                        x2_mirrored[1],
                                                                                                                                        path_length[0] / units.m))
            tmp += path_length[0]
        return tmp

    def get_path_length_analytic(self, x1, x2, C_0, reflection=0, reflection_case=1):
        """
        analytic solution to calculate the distance along the path. This code is based on the analytic solution found
        by Ben Hokanson-Fasing and the pyrex implementation.
        """
        
        tmp = 0
        for iS, segment in enumerate(self.get_path_segments(x1, x2, C_0, reflection, reflection_case)):
            if(iS == 0 and reflection_case == 2):  # we can only integrate upward going rays, so if the ray starts downwardgoing, we need to mirror
                x11, x1, x22, x2, C_0, C_1 = segment
                x1t = copy.copy(x11)
                x2t = copy.copy(x2)
                x1t[1] = x2[1]
                x2t[1] = x11[1]
                x2 = x2t
                x1 = x1t
            else:
                x11, x1, x22, x2, C_0, C_1 = segment
        
            solution_type = self.determine_solution_type(x1, x2, C_0)
    
            z_deep = get_z_deep((self.medium.n_ice, self.medium.z_0, self.medium.delta_n))
            launch_angle = self.get_launch_angle(x1, C_0)
            beta = self.n(x1[1]) * np.sin(launch_angle)
            alpha = self.medium.n_ice ** 2 - beta ** 2
    #         print("launchangle {:.1f} beta {:.2g} alpha {:.2g}, n(z1) = {:.2g} n(z2) = {:.2g}".format(launch_angle/units.deg, beta, alpha, self.n(x1[1]), self.n(x2[1])))
    
            def l1(z):
                gamma = self.n(z) ** 2 - beta ** 2
                gamma = np.where(gamma < 0, 0, gamma)
                return self.medium.n_ice * self.n(z) - beta ** 2 - (alpha * gamma) ** 0.5
    
            def l2(z):
                gamma = self.n(z) ** 2 - beta ** 2
                gamma = np.where(gamma < 0, 0, gamma)
                return self.n(z) + gamma ** 0.5
    
            def get_s(z, deep=False):
                if(deep):
                    return self.medium.n_ice * z / alpha ** 0.5
                else:
                    #                 print(z, self.n(z), beta)
                    #                 print(alpha**0.5, l1(z), l2(z))
    
                    path_length = self.medium.n_ice / alpha ** 0.5 * (-z + np.log(l1(z)) * self.medium.z_0) + np.log(l2(z)) * self.medium.z_0
                    if (np.abs(path_length) == np.inf or path_length == np.nan):
                        path_length = None
    
                    return path_length
    
            def get_path_direct(z1, z2):
                int1 = get_s(z1, z1 < z_deep)
                int2 = get_s(z2, z2 < z_deep)
                if (int1 == None or int2 == None):
                    return None
    #             print('analytic {:.4g} ({:.0f} - {:.0f}={:.4g}, {:.4g})'.format(
    #                 int2 - int1, get_s(x2[1]), x1[1], x2[1], get_s(x1[1])))
                if (z1 < z_deep) == (z2 < z_deep):
                    # z0 and z1 on same side of z_deep
                    return int2 - int1
                else:
                    try:
                        int_diff = get_s(z_deep, deep=True) - get_s(z_deep, deep=False)
                    except:
                        return None
                    if z1 < z2:
                        # z0 below z_deep, z1 above z_deep
                        return int2 - int1 + int_diff
                    else:
                        # print("path:", int2 - int1 - int_diff)
                        # z0 above z_deep, z1 below z_deep
                        return int2 - int1 - int_diff
    
            if(solution_type == 1):
                tmp += get_path_direct(x1[1], x2[1])
            else:
                if(solution_type == 3):
                    z_turn = 0
                else:
                    gamma_turn, z_turn = self.get_turning_point(self.medium.n_ice ** 2 - C_0 ** -2)
    #             print('solution type {:d}, zturn = {:.1f}'.format(solution_type, z_turn))
                try:
                    tmp += get_path_direct(x1[1], z_turn) + get_path_direct(x2[1], z_turn)
                except:
                    tmp += None
            
        return tmp
    
    def get_travel_time(self, x1, x2, C_0, reflection=0, reflection_case=1):
        tmp = 0
        for iS, segment in enumerate(self.get_path_segments(x1, x2, C_0, reflection, reflection_case)):
            if(iS == 0 and reflection_case == 2):  # we can only integrate upward going rays, so if the ray starts downwardgoing, we need to mirror
                x11, x1, x22, x2, C_0, C_1 = segment
                x1t = copy.copy(x11)
                x2t = copy.copy(x2)
                x1t[1] = x2[1]
                x2t[1] = x11[1]
                x2 = x2t
                x1 = x1t
            else:
                x11, x1, x22, x2, C_0, C_1 = segment
        
            x2_mirrored = self.get_z_mirrored(x1, x2, C_0)
    
            def dt(t, C_0):
                z = self.get_z_unmirrored(t, C_0)
                return self.ds(t, C_0) / speed_of_light * self.n(z)
    
            gamma_turn, z_turn = self.get_turning_point(self.medium.n_ice ** 2 - C_0 ** -2)
            points = None
            if(x1[1] < z_turn and z_turn < x2_mirrored[1]):
                points = [z_turn]
            travel_time = integrate.quad(dt, x1[1], x2_mirrored[1], args=(C_0), points=points, epsabs=1e-10, epsrel=1.49e-08, limit=500)
            self.__logger.info("calculating travel time from ({:.0f}, {:.0f}) to ({:.0f}, {:.0f}) = ({:.0f}, {:.0f}) = {:.2f} ns".format(
                x1[0], x1[1], x2[0], x2[1], x2_mirrored[0], x2_mirrored[1], travel_time[0] / units.ns))
            tmp += travel_time[0]
        return tmp

    def get_travel_time_analytic(self, x1, x2, C_0, reflection=0, reflection_case=1):
        """
        analytic solution to calculate the time of flight. This code is based on the analytic solution found
        by Ben Hokanson-Fasing and the pyrex implementation.
        """
        
        tmp = 0
        for iS, segment in enumerate(self.get_path_segments(x1, x2, C_0, reflection, reflection_case)):
            if(iS == 0 and reflection_case == 2):  # we can only integrate upward going rays, so if the ray starts downwardgoing, we need to mirror
                x11, x1, x22, x2, C_0, C_1 = segment
                x1t = copy.copy(x11)
                x2t = copy.copy(x2)
                x1t[1] = x2[1]
                x2t[1] = x11[1]
                x2 = x2t
                x1 = x1t
            else:
                x11, x1, x22, x2, C_0, C_1 = segment
        
            solution_type = self.determine_solution_type(x1, x2, C_0)
    
            z_deep = get_z_deep((self.medium.n_ice, self.medium.z_0, self.medium.delta_n))
            launch_angle = self.get_launch_angle(x1, C_0)
            beta = self.n(x1[1]) * np.sin(launch_angle)
            alpha = self.medium.n_ice ** 2 - beta ** 2
    #         print("launchangle {:.1f} beta {:.2g} alpha {:.2g}, n(z1) = {:.2g} n(z2) = {:.2g}".format(launch_angle/units.deg, beta, alpha, self.n(x1[1]), self.n(x2[1])))
    
            def l1(z):
                gamma = self.n(z) ** 2 - beta ** 2
                gamma = np.where(gamma < 0, 0, gamma)
                return self.medium.n_ice * self.n(z) - beta ** 2 - (alpha * gamma) ** 0.5
    
            def l2(z):
                gamma = self.n(z) ** 2 - beta ** 2
                gamma = np.where(gamma < 0, 0, gamma)
                return self.n(z) + gamma ** 0.5
    
            def get_s(z, deep=False):
                if(deep):
                    return self.medium.n_ice * (self.n(z) + self.medium.n_ice * (z / self.medium.z_0 - 1)) / (np.sqrt(alpha) / self.medium.z_0 * speed_of_light)
                else:
                    gamma = self.n(z) ** 2 - beta ** 2
                    gamma = np.where(gamma < 0, 0, gamma)
                    log_1 = l1(z)
                    log_2 = l2(z)
                    s = (((np.sqrt(gamma) + self.medium.n_ice * np.log(log_2) + 
                              self.medium.n_ice ** 2 * np.log(log_1) / np.sqrt(alpha)) * self.medium.z_0) - 
                            z * self.medium.n_ice ** 2 / np.sqrt(alpha)) / speed_of_light
                    if (np.abs(s) == np.inf or s == np.nan):
                        s = None
    
                    return s
    
            def get_ToF_direct(z1, z2):
                int1 = get_s(z1, z1 < z_deep)
                int2 = get_s(z2, z2 < z_deep)
                if (int1 == None or int2 == None):
                    return None
    #             print('analytic {:.4g} ({:.0f} - {:.0f}={:.4g}, {:.4g})'.format(
    #                 int2 - int1, get_s(x2[1]), x1[1], x2[1], get_s(x1[1])))
                if (z1 < z_deep) == (z2 < z_deep):
                    # z0 and z1 on same side of z_deep
                    return int2 - int1
                else:
                    try:
                        int_diff = get_s(z_deep, deep=True) - get_s(z_deep, deep=False)
                    except:
                        return None
                    if z1 < z2:
                        # z0 below z_deep, z1 above z_deep
                        return int2 - int1 + int_diff
                    else:
                        # z0 above z_deep, z1 below z_deep
                        return int2 - int1 - int_diff
    
            if(solution_type == 1):
                ttmp = get_ToF_direct(x1[1], x2[1])
                tmp += ttmp
                self.__logger.info("calculating travel time from ({:.0f}, {:.0f}) to ({:.0f}, {:.0f}) = {:.2f} ns".format(
                    x1[0], x1[1], x2[0], x2[1], ttmp / units.ns))
            else:
                if(solution_type == 3):
                    z_turn = 0
                else:
                    gamma_turn, z_turn = self.get_turning_point(self.medium.n_ice ** 2 - C_0 ** -2)
    #             print('solution type {:d}, zturn = {:.1f}'.format(solution_type, z_turn))
                try:
                    ttmp = get_ToF_direct(x1[1], z_turn) + get_ToF_direct(x2[1], z_turn)
                    tmp += ttmp
                    self.__logger.info("calculating travel time from ({:.0f}, {:.0f}) to ({:.0f}, {:.0f}) = {:.2f} ns".format(
                        x1[0], x1[1], x2[0], x2[1], ttmp / units.ns))
                except:
                    tmp += None
        return tmp

    def __get_frequencies_for_attenuation(self, frequency, max_detector_freq):
            mask = frequency > 0
            nfreqs = min(self.__n_frequencies_integration, np.sum(mask))
            freqs = np.linspace(frequency[mask].min(), frequency[mask].max(), nfreqs)
            if(nfreqs < np.sum(mask) and max_detector_freq is not None):
                mask2 = frequency <= max_detector_freq
                nfreqs2 = min(self.__n_frequencies_integration, np.sum(mask2 & mask))
                freqs = np.linspace(frequency[mask2 & mask].min(), frequency[mask2 & mask].max(), nfreqs2)
                if(np.sum(~mask2) > 1):
                    freqs = np.append(freqs, np.linspace(frequency[~mask2].min(), frequency[~mask2].max(), nfreqs // 2))
            return freqs

    def get_attenuation_along_path(self, x1, x2, C_0, frequency, max_detector_freq, reflection=0, reflection_case=1):
        tmp_attenuation = None
        for iS, segment in enumerate(self.get_path_segments(x1, x2, C_0, reflection, reflection_case)):
            if(iS == 0 and reflection_case == 2):  # we can only integrate upward going rays, so if the ray starts downwardgoing, we need to mirror
                x11, x1, x22, x2, C_0, C_1 = segment
                x1t = copy.copy(x11)
                x2t = copy.copy(x2)
                x1t[1] = x2[1]
                x2t[1] = x11[1]
                x2 = x2t
                x1 = x1t
            else:
                x11, x1, x22, x2, C_0, C_1 = segment

            if(cpp_available):
                mask = frequency > 0
                freqs = self.__get_frequencies_for_attenuation(frequency, max_detector_freq)
                tmp = np.zeros_like(freqs)
                for i, f in enumerate(freqs):
                    tmp[i] = wrapper.get_attenuation_along_path(
                        x1, x2, C_0, f, self.medium.n_ice, self.medium.delta_n, self.medium.z_0, self.attenuation_model_int)
                attenuation = np.ones_like(frequency)
                attenuation[mask] = np.interp(frequency[mask], freqs, tmp)
            else:
    
                x2_mirrored = self.get_z_mirrored(x1, x2, C_0)
    
                def dt(t, C_0, frequency):
                    z = self.get_z_unmirrored(t, C_0)
                    return self.ds(t, C_0) / attenuation_util.get_attenuation_length(z, frequency, self.attenuation_model)
    
                # to speed up things we only calculate the attenuation for a few frequencies
                # and interpolate linearly between them
                mask = frequency > 0
                freqs = self.__get_frequencies_for_attenuation(frequency, max_detector_freq)
                gamma_turn, z_turn = self.get_turning_point(self.medium.n_ice ** 2 - C_0 ** -2)
                points = None
                if(x1[1] < z_turn and z_turn < x2_mirrored[1]):
                    points = [z_turn]
                tmp = np.array([integrate.quad(dt, x1[1], x2_mirrored[1], args=(
                    C_0, f), epsrel=1e-2, points=points)[0] for f in freqs])
                tmp = np.exp(-1 * tmp)
        #         tmp = np.array([integrate.quad(dt, x1[1], x2_mirrored[1], args=(C_0, f), epsrel=0.05)[0] for f in frequency[mask]])
                attenuation = np.ones_like(frequency)
                attenuation[mask] = np.interp(frequency[mask], freqs, tmp)
                self.__logger.info("calculating attenuation from ({:.0f}, {:.0f}) to ({:.0f}, {:.0f}) = ({:.0f}, {:.0f}) =  a factor {}".format(
                    x1[0], x1[1], x2[0], x2[1], x2_mirrored[0], x2_mirrored[1], 1 / attenuation))
            if(tmp_attenuation is None):
                tmp_attenuation = attenuation
            else:
                tmp_attenuation *= attenuation
        return attenuation

    def get_path_segments(self, x1, x2, C_0, reflection=0, reflection_case=1):
        """
        Calculates the different segments of the path that makes up the full ray tracing path
        One segment per bottom reflection. 
        
        Parameters
        ----------
        x1: tuple
            (y, z) coordinate of start value
        x2: tuple
            (y, z) coordinate of stop value
        C_0: float
            C_0 parameter of analytic ray path function
        reflection: int (default 0)
            the number of bottom reflections to consider
        reflection_case: int (default 1)
            only relevant if `reflection` is larger than 0
            * 1: rays start upwards
            * 2: rays start downwards
            
        Returns
        --------
        (original x1, x1 of path segment, original x2, x2 of path segment, C_0, C_1 of path segment)
        """
        x1 = copy.copy(x1)
        x11 = copy.copy(x1)
        x22 = copy.copy(x2)
        
        if(reflection == 0):
            C_1 = self.get_C_1(x1, C_0)
            return [[x1, x1, x22, x2, C_0, C_1]]
        
        tmp = []
        
        if(reflection_case == 2):
            # the code only allows upward going rays, thus we find a point left from x1 that has an upward going ray
            # that will produce a downward going ray through x1
            y_turn = self.get_y_turn(C_0, x1)
            dy = y_turn - x1[0]
            self.__logger.debug("relaction case 2: shifting x1 {} to {}".format(x1, x1[0] - 2 * dy))
            x1[0] = x1[0] - 2 * dy
        
        for i in range(reflection + 1):
            self.__logger.debug("calculation path for reflection = {}".format(i))
            C_1 = self.get_C_1(x1, C_0)
            x2 = self.get_reflection_point(C_0, C_1)
            stop_loop = False
            if(x2[0] > x22[0]):
                stop_loop = True
                x2 = x22
            tmp.append([x11, x1, x22, x2, C_0, C_1])
            if(stop_loop):
                break
#             yyy, zzz = self.get_path(x1, x2, C_0, n_points)
#             yy.extend(yyy)
#             zz.extend(zzz)
            self.__logger.debug("setting x1 from {} to {}".format(x1, x2))
            x1 = x2
        return tmp
    
    def get_angle(self, x, x_start, C_0, reflection=0, reflection_case=1):
        """
        calculates the angle with respect to the positive z-axis of the ray path at position x
        
        Parameters
        ----------
        x: tuple
            (y, z) coordinate to calculate the angle
        x_start: tuple
            (y, z) start position of the ray
        C_0: float
            C_0 parameter of analytic ray path function
        reflection: int (default 0)
            the number of bottom reflections to consider
        reflection_case: int (default 1)
            only relevant if `reflection` is larger than 0
            * 1: rays start upwards
            * 2: rays start downwards
        """
        last_segment = self.get_path_segments(x_start, x, C_0, reflection, reflection_case)[-1]
        x_start = last_segment[1]
            
        z = self.get_z_mirrored(x_start, x, C_0)[1]
        dy = self.get_y_diff(z, C_0)
        angle = np.arctan(dy)
        if(angle < 0):
            angle = np.pi + angle
        return angle

    def get_launch_angle(self, x1, C_0, reflection=0, reflection_case=1):
        return self.get_angle(x1, x1, C_0, reflection, reflection_case)

    def get_receive_angle(self, x1, x2, C_0, reflection=0, reflection_case=1):
        return np.pi - self.get_angle(x2, x1, C_0, reflection, reflection_case)

    def get_reflection_angle(self, x1, x2, C_0, reflection=0, reflection_case=1):
        """
        calculates the angle under which the ray reflects off the surface. If not reflection occurs, None is returned
        
        If reflections off the bottom (e.g. Moore's Bay) are simulated, an array with reflection angles (one for 
        each track segment) is returned
        
        Parameters
        ----------
        x1: tuple
            (y, z) start position of ray
        x2: tuple
            (y, z) stop position of the ray
        C_0: float
            C_0 parameter of analytic ray path function
        reflection: int (default 0)
            the number of bottom reflections to consider
        reflection_case: int (default 1)
            only relevant if `reflection` is larger than 0
            * 1: rays start upwards
            * 2: rays start downwards
        """
        output = []
        c = self.medium.n_ice ** 2 - C_0 ** -2
        for segment in self.get_path_segments(x1, x2, C_0, reflection, reflection_case):
            x11, x1, x22, x2, C_0, C_1 = segment
            gamma_turn, z_turn = self.get_turning_point(c)
            y_turn = self.get_y_turn(C_0, x1)
            if((z_turn >= 0) and (y_turn > x11[0]) and (y_turn < x22[0])):  # for the first track segment we need to check if turning point is right of start point (otherwise we have a downward going ray that does not have a turning point), and for the last track segment we need to check that the turning point is left of the stop position. 
                r = self.get_angle(np.array([y_turn, 0]), x1, C_0)
                self.__logger.debug(
                    "reflecting off surface at y = {:.1f}m, reflection angle = {:.1f}deg".format(y_turn, r / units.deg))
                output.append(r)
            else:
                output.append(None)
        return np.squeeze(output)

    def get_path(self, x1, x2, C_0, n_points=1000):
        """
        for plotting purposes only, returns the ray tracing path between x1 and x2

        the result is only valid if C_0 is a solution to the ray tracing problem

        Parameters
        -------
        x1: array
            start position (y, z)
        x2: array
            stop position (y, z)
        C_0: (float)
            first parameter
        n_points: integer (optional)
            the number of coordinates to calculate

        Returns
        -------
        yy: array
            the y coordinates of the ray tracing path
        zz: array
            the z coordinates of the ray tracing path
        """
        c = self.medium.n_ice ** 2 - C_0 ** -2
        C_1 = x1[0] - self.get_y_with_z_mirror(x1[1], C_0)
        gamma_turn, z_turn = self.get_turning_point(c)
        y_turn = self.get_y(gamma_turn, C_0, C_1)
        zstart = x1[1]
        zstop = self.get_z_mirrored(x1, x2, C_0)[1]
        z = np.linspace(zstart, zstop, n_points)
        mask = z < z_turn
        res = np.zeros_like(z)
        zs = np.zeros_like(z)
        gamma = self.get_gamma(z[mask])
        zs[mask] = z[mask]
        res[mask] = self.get_y(gamma, C_0, C_1)
        gamma = self.get_gamma(2 * z_turn - z[~mask])
        res[~mask] = 2 * y_turn - self.get_y(gamma, C_0, C_1)
        zs[~mask] = 2 * z_turn - z[~mask]

        self.__logger.debug('turning points for C_0 = {:.2f}, b= {:.2f}, gamma = {:.4f}, z = {:.1f}, y_turn = {:.0f}'.format(
            C_0, self.__b, gamma_turn, z_turn, y_turn))
        return res, zs
    
    def get_path_reflections(self, x1, x2, C_0, n_points=1000, reflection=0, reflection_case=1):
        """
        calculates the ray path in the presence of reflections at the bottom
        The full path is constructed by multiple calls to the `get_path()` function to put together the full path
        
        Parameters
        ----------
        x1: tuple
            (y, z) coordinate of start value
        x2: tuple
            (y, z) coordinate of stop value
        C_0: float
            C_0 parameter of analytic ray path function
        n_points: int (default 1000)
            the number of points of the numeric path
        reflection: int (default 0)
            the number of bottom reflections to consider
        reflection_case: int (default 1)
            only relevant if `reflection` is larger than 0
            * 1: rays start upwards
            * 2: rays start downwards
        
        Returns
        -------
        yy: array
            the y coordinates of the ray tracing path
        zz: array
            the z coordinates of the ray tracing path
        """
        yy = []
        zz = []
        x1 = copy.copy(x1)
        x11 = copy.copy(x1)
        
        if(reflection and reflection_case == 2):
            # the code only allows upward going rays, thus we find a point left from x1 that has an upward going ray
            # that will produce a downward going ray through x1
            y_turn = self.get_y_turn(C_0, x1)
            dy = y_turn - x1[0]
            self.__logger.debug("relaction case 2: shifting x1 {} to {}".format(x1, x1[0] - 2 * dy))
            x1[0] = x1[0] - 2 * dy
        
        x22 = copy.copy(x2)
        for i in range(reflection + 1):
            self.__logger.debug("calculation path for reflection = {}".format(i))
            C_1 = x1[0] - self.get_y_with_z_mirror(x1[1], C_0)
            x2 = self.get_reflection_point(C_0, C_1)
            if(x2[0] > x22[0]):
                x2 = x22
            yyy, zzz = self.get_path(x1, x2, C_0, n_points)
            yy.extend(yyy)
            zz.extend(zzz)
            self.__logger.debug("setting x1 from {} to {}".format(x1, x2))
            x1 = x2

        yy = np.array(yy)
        zz = np.array(zz)
        mask = yy > x11[0]
        return yy[mask], zz[mask]
            
    def get_reflection_point(self, C_0, C_1):
        """
        calculates the point where the signal gets reflected off the bottom of the ice shelf
        
        Returns tuple (y,z)
        """
        c = self.medium.n_ice ** 2 - C_0 ** -2
        gamma_turn, z_turn = self.get_turning_point(c)
        x2 = [0, self.medium.reflection]
        x2[0] = self.get_y_with_z_mirror(-x2[1] + 2 * z_turn, C_0, C_1)
        return x2

    def obj_delta_y_square(self, logC_0, x1, x2, reflection=0, reflection_case=2):
        """
        objective function to find solution for C0
        """
        C_0 = self.get_C0_from_log(logC_0)
        return self.get_delta_y(C_0, copy.copy(x1), x2, reflection=reflection, reflection_case=reflection_case) ** 2

    def obj_delta_y(self, logC_0, x1, x2, reflection=0, reflection_case=2):
        """
        function to find solution for C0, returns distance in y between function and x2 position
        result is signed! (important to use a root finder)
        """
        C_0 = self.get_C0_from_log(logC_0)
        return self.get_delta_y(C_0, copy.copy(x1), x2, reflection=reflection, reflection_case=reflection_case)

    def get_delta_y(self, C_0, x1, x2, C0range=None, reflection=0, reflection_case=2):
        """
        calculates the difference in the y position between the analytic ray tracing path
        specified by C_0 at the position x2
        """
        if(C0range is None):
            C0range = [1. / self.medium.n_ice, np.inf]
        if(hasattr(C_0, '__len__')):
            C_0 = C_0[0]
        if((C_0 < C0range[0]) or(C_0 > C0range[1])):
            self.__logger.debug('C0 = {:.4f} out of range {:.0f} - {:.2f}'.format(C_0, C0range[0], C0range[1]))
            return -np.inf
        c = self.medium.n_ice ** 2 - C_0 ** -2
        
        # we consider two cases here, 
        # 1) the rays start rising -> the default case
        # 2) the rays start decreasing -> we need to find the position left of the start point that
        #    that has rising rays that go through the point x1
        if(reflection > 0 and reflection_case == 2):
            y_turn = self.get_y_turn(C_0, x1)
            dy = y_turn - x1[0]
            self.__logger.debug("relaction case 2: shifting x1 {} to {}".format(x1, x1[0] - 2 * dy))
            x1[0] = x1[0] - 2 * dy

        for i in range(reflection):
            # we take account reflections at the bottom layer into account via
            # 1) calculating the point where the reflection happens
            # 2) starting a ray tracing from this new point
            
            # determine y translation first
            C_1 = x1[0] - self.get_y_with_z_mirror(x1[1], C_0)
            if(hasattr(C_1, '__len__')):
                C_1 = C_1[0]
    
            self.__logger.debug("C_0 = {:.4f}, C_1 = {:.1f}".format(C_0, C_1))
            x1 = self.get_reflection_point(C_0, C_1)

        # determine y translation first
        C_1 = x1[0] - self.get_y_with_z_mirror(x1[1], C_0)
        if(hasattr(C_1, '__len__')):
            C_1 = C_1[0]

        self.__logger.debug("C_0 = {:.4f}, C_1 = {:.1f}".format(C_0, C_1))

        # for a given c_0, 3 cases are possible to reach the y position of x2
        # 1) direct ray, i.e., before the turning point
        # 2) refracted ray, i.e. after the turning point but not touching the surface
        # 3) reflected ray, i.e. after the ray reaches the surface
        gamma_turn, z_turn = self.get_turning_point(c)
        y_turn = self.get_y(gamma_turn, C_0, C_1)
        if(z_turn < x2[1]):  # turning points is deeper that x2 positions, can't reach target
            # the minimizer has problems finding the minimum if inf is returned here. Therefore, we return the distance
            # between the turning point and the target point + 10 x the distance between the z position of the turning points
            # and the target position. This results in a objective function that has the solutions as the only minima and
            # is smooth in C_0
            diff = ((z_turn - x2[1]) ** 2 + (y_turn - x2[0]) ** 2) ** 0.5 + 10 * np.abs(z_turn - x2[1])
            self.__logger.debug(
                "turning points (zturn = {:.0f} is deeper than x2 positon z2 = {:.0f}, setting distance to target position to {:.1f}".format(z_turn, x2[1], -diff))
            return -diff
#             return -np.inf
        self.__logger.debug('turning points is z = {:.1f}, y =  {:.1f}'.format(z_turn, y_turn))
        if(y_turn > x2[0]):  # we always propagate from left to right
            # direct ray
            y2_fit = self.get_y(self.get_gamma(x2[1]), C_0, C_1)  # calculate y position at get_path position
            diff = (x2[0] - y2_fit)
            if(hasattr(diff, '__len__')):
                diff = diff[0]
            if(hasattr(x2[0], '__len__')):
                x2[0] = x2[0][0]

            self.__logger.debug(
                'we have a direct ray, y({:.1f}) = {:.1f} -> {:.1f} away from {:.1f}, turning point = y={:.1f}, z={:.2f}, x0 = {:.1f} {:.1f}'.format(x2[1], y2_fit, diff, x2[0], y_turn, z_turn, x1[0], x1[1]))
            return diff
        else:
            # now it's a bit more complicated. we need to transform the coordinates to
            # be on the mirrored part of the function
            z_mirrored = x2[1]
            gamma = self.get_gamma(z_mirrored)
            self.__logger.debug("get_y( {}, {}, {})".format(gamma, C_0, C_1))
            y2_raw = self.get_y(gamma, C_0, C_1)
            y2_fit = 2 * y_turn - y2_raw
            diff = (x2[0] - y2_fit)

            self.__logger.debug('we have a reflected/refracted ray, y({:.1f}) = {:.1f} ({:.1f}) -> {:.1f} away from {:.1f} (gamma = {:.5g})'.format(
                z_mirrored, y2_fit, y2_raw, diff, x2[0], gamma))
            return -1 * diff

    def determine_solution_type(self, x1, x2, C_0):
        """ returns the type of the solution

        Parameters
        ----------
        x1: 2dim np.array
            start position
        x2: 2dim np.array
            stop position
        C_0: float
            C_0 value of ray tracing solution

        Returns
        -------
        solution_type: int
            * 1: 'direct'
            * 2: 'refracted'
            * 3: 'reflected
        """
        c = self.medium.n_ice ** 2 - C_0 ** -2
        C_1 = x1[0] - self.get_y_with_z_mirror(x1[1], C_0)
        gamma_turn, z_turn = self.get_turning_point(c)
        y_turn = self.get_y(gamma_turn, C_0, C_1)
        if(x2[0] < y_turn):
            return 1
        else:
            if(z_turn == 0):
                return 3
            else:
                return 2

    def find_solutions(self, x1, x2, plot=False, reflection=0, reflection_case=1):
        """
        this function finds all ray tracing solutions

        prerequesite is that x2 is above and to the right of x1, this is not a violation of universality
        because this requirement can be achieved with a simple coordinate transformation

        Parameters
        -----------
        x1: tuple
            (y,z) coordinate of start point
        x2: tuple
            (y,z) coordinate of stop point
        reflection: int (default 0)
            how many reflections off the reflective layer (bottom of ice shelf) should be simulated
        

        returns an array of the C_0 paramters of the solutions (the array might be empty)
        
        """
        
        if(reflection > 0 and self.medium.reflection is None):
            self.__logger.error("a solution for {:d} reflection(s) off the bottom reflective layer is requested, but ice model does not specify a reflective layer".format(reflection))
            raise AttributeError("a solution for {:d} reflection(s) off the bottom reflective layer is requested, but ice model does not specify a reflective layer".format(reflection))

        if(cpp_available):
            #             t = time.time()
#             print("find solutions", x1, x2, self.medium.n_ice, self.medium.delta_n, self.medium.z_0, reflection, reflection_case, self.medium.reflection)
            tmp_reflection = copy.copy(self.medium.reflection)
            if(tmp_reflection is None):
                tmp_reflection = 100  # this parameter will never be used but is required to be an into to be able to pass it to the C++ module, so set it to a positive number, i.e., a reflective layer above the ice
            solutions = wrapper.find_solutions(x1, x2, self.medium.n_ice, self.medium.delta_n, self.medium.z_0, reflection, reflection_case, tmp_reflection)
#             print((time.time() -t)*1000.)
            return solutions
        else:

            tol = 1e-6
            results = []
            C0s = []  # intermediate storage of results

            # calculate optimal start value. The objective function becomes infinity if the turning point is below the z
            # position of the observer. We calculate the corresponding value so that the minimization starts at one edge
            # of the objective function
            # c = self.__b ** 2 / 4 - (0.5 * self.__b - np.exp(x2[1] / self.medium.z_0) * self.medium.n_ice) ** 2
            # C_0_start = (1 / (self.medium.n_ice ** 2 - c)) ** 0.5
            # R.L. March 15, 2019: This initial condition does not find a solution for e.g.:
            # emitter  at [-400.0*units.m,-732.0*units.m], receiver at [0., -2.0*units.m]

            if(self.__use_optimized_start_values):
                # take surface skimming ray as start value
                C_0_start, th_start = self.get_surf_skim_angle(x1)
                logC_0_start = np.log(C_0_start - 1. / self.medium.n_ice)
                self.__logger.debug(
                    'starting optimization with x0 = {:.2f} -> C0 = {:.3f}'.format(logC_0_start, C_0_start))
            else:
                logC_0_start = -1

            result = optimize.root(self.obj_delta_y_square, x0=logC_0_start, args=(x1, x2, reflection, reflection_case), tol=tol)

            if(plot):
                import matplotlib.pyplot as plt
                fig, ax = plt.subplots(1, 1)
            if(result.fun < 1e-7):
                if(plot):
                    self.plot_result(x1, x2, self.get_C0_from_log(result.x[0]), ax)
                if(np.round(result.x[0], 3) not in np.round(C0s, 3)):
                    C_0 = self.get_C0_from_log(result.x[0])
                    C0s.append(C_0)
                    solution_type = self.determine_solution_type(x1, x2, C_0)
                    self.__logger.info("found {} solution C0 = {:.2f}".format(solution_types[solution_type], C_0))
                    results.append({'type': solution_type,
                                    'C0': C_0,
                                    'C1': self.get_C_1(x1, C_0),
                                    'reflection': reflection,
                                    'reflection_case': reflection_case})

            # check if another solution with higher logC0 exists
            logC0_start = result.x[0] + 0.0001
            logC0_stop = 100
            delta_start = self.obj_delta_y(logC0_start, x1, x2, reflection, reflection_case)
            delta_stop = self.obj_delta_y(logC0_stop, x1, x2, reflection, reflection_case)
        #     print(logC0_start, logC0_stop, delta_start, delta_stop, np.sign(delta_start), np.sign(delta_stop))
            if(np.sign(delta_start) != np.sign(delta_stop)):
                self.__logger.info("solution with logC0 > {:.3f} exists".format(result.x[0]))
                result2 = optimize.brentq(self.obj_delta_y, logC0_start, logC0_stop, args=(x1, x2, reflection, reflection_case))
                if(plot):
                    self.plot_result(x1, x2, self.get_C0_from_log(result2), ax)
                if(np.round(result2, 3) not in np.round(C0s, 3)):
                    C_0 = self.get_C0_from_log(result2)
                    C0s.append(C_0)
                    solution_type = self.determine_solution_type(x1, x2, C_0)
                    self.__logger.info("found {} solution C0 = {:.2f}".format(solution_types[solution_type], C_0))
                    results.append({'type': solution_type,
                                    'C0': C_0,
                                    'C1': self.get_C_1(x1, C_0),
                                    'reflection': reflection,
                                    'reflection_case': reflection_case})
            else:
                self.__logger.info("no solution with logC0 > {:.3f} exists".format(result.x[0]))

            logC0_start = -100
            logC0_stop = result.x[0] - 0.0001
            delta_start = self.obj_delta_y(logC0_start, x1, x2, reflection, reflection_case)
            delta_stop = self.obj_delta_y(logC0_stop, x1, x2, reflection, reflection_case)
        #     print(logC0_start, logC0_stop, delta_start, delta_stop, np.sign(delta_start), np.sign(delta_stop))
            if(np.sign(delta_start) != np.sign(delta_stop)):
                self.__logger.info("solution with logC0 < {:.3f} exists".format(result.x[0]))
                result3 = optimize.brentq(self.obj_delta_y, logC0_start, logC0_stop, args=(x1, x2, reflection, reflection_case))

                if(plot):
                    self.plot_result(x1, x2, self.get_C0_from_log(result3), ax)
                if(np.round(result3, 3) not in np.round(C0s, 3)):
                    C_0 = self.get_C0_from_log(result3)
                    C0s.append(C_0)
                    solution_type = self.determine_solution_type(x1, x2, C_0)
                    self.__logger.info("found {} solution C0 = {:.2f}".format(solution_types[solution_type], C_0))
                    results.append({'type': solution_type,
                                    'C0': C_0,
                                    'C1': self.get_C_1(x1, C_0),
                                    'reflection': reflection,
                                    'reflection_case': reflection_case})
            else:
                self.__logger.info("no solution with logC0 < {:.3f} exists".format(result.x[0]))

            if(plot):
                import matplotlib.pyplot as plt
                plt.show()

            return sorted(results, key=itemgetter('type'))

    def plot_result(self, x1, x2, C_0, ax):
        """
        helper function to visualize results
        """
        C_1 = self.get_C_1(x1, C_0)

        zs = np.linspace(x1[1], x1[1] + np.abs(x1[1]) + np.abs(x2[1]), 1000)
        yy, zz = self.get_y_with_z_mirror(zs, C_0, C_1)
        ax.plot(yy, zz, '-', label='C0 = {:.3f}'.format(C_0))
        ax.plot(x1[0], x1[1], 'ko')
        ax.plot(x2[0], x2[1], 'd')

    #     ax.plot(zz, yy, '-', label='C0 = {:.3f}'.format(C_0))
    #     ax.plot(x1[1], x1[0], 'ko')
    #     ax.plot(x2[1], x2[0], 'd')
        ax.legend()

    def get_angle_from_C_0(self, C_0, z_pos, angoff=0):
        logC_0 = np.log(C_0 - 1. / self.medium.n_ice)
        return self.get_angle_from_logC_0(logC_0, z_pos, angoff)

    def get_angle_from_logC_0(self, logC_0, z_pos, angoff=0):

        '''
        argument angoff is provided so that the function can be used for minimization in get_C_0_from_angle(),
        in which case angoff is the angle for which the C_0 is sought and zero is returned when it is found.

        C_0 has a smallest possible value at 1./self.medium.n_ice . When it approaches this value, very
        small changes in C_0 correspond to a given change in the angle. In order to prevent the root finding
        algorithm from crossing into the invalid range of C_0 at values smaller than 1./self.medium.n_ice,
        the root finding is done with the parameter logC_0 = np.log(C_0 - 1. / self.medium.n_ice), so it is
        not exactly the log of C_0 as the nome of this method implies.
        This is the same parameter transformation that is done for find_solutions()

        input:
            logC_0 = np.log(C_0 - 1. / self.medium.n_ice)
            angoff = angular offset
            z_pos  = z-position from where ray is emitted
        output:
            angle corresponding to C_0, minus offset angoff
        '''

        C_0 = self.get_C0_from_log(logC_0)

        dydz = self.get_y_diff(z_pos, C_0)
#        dydz = self.get_dydz_analytic(C_0, z_pos)
        angle = np.arctan(dydz)

        # print(dydz,angoffdydz)

        return angle - angoff

    def get_C_0_from_angle(self, anglaunch, z_pos):

        '''
        Find parameter C0 corresponding to a given launch angle and z-position of a ray.
        The parameter is found by means of a root finding procedure

        output:
            Complete output of optimisation procedure
            (result.x[0] is the C0 value found by optimisation procedure)

        '''

        # C_0 has a smallest possible value at 1./self.medium.n_ice . When it approaches this value, very
        # small changes in C_0 correspond to given change in the angle. In order to prevent the root finding
        # algorithm to cross into the invalid range of C_0 at  values smaller than 1./self.medium.n_ice,
        # the root finding is done with the parameter logC_0_start below. This is the same parameter transformation
        # that is done for find_solutions()

        C_0_start = 2.

        logC_0_start = np.log(C_0_start - 1. / self.medium.n_ice)

#        result = optimize.root(self.get_angle_from_C_0,np.pi/4.,args=(z_pos,anglaunch))
        result = optimize.root(self.get_angle_from_logC_0, logC_0_start, args=(z_pos, anglaunch))

        # want to return the complete instance of the result class; result value result.x[0] is logC_0,
        # but we want C_0, so replace it in the result class. This may not be good practice but it seems to be
        # more user-friendly than to return the value logC_0
        result.x[0] = copy.copy(self.get_C0_from_log(result.x[0]))

        return result

#     def get_dydz_analytic(self, C_0, z_pos):
#         '''
#         Implementation of derivative dy/dz obtained from the analytic expresion for y(z)
#         Returns dy/dz for a given z-position and C_0
#         '''
#
#         gamma = self.get_gamma(z_pos)
#
#         b = self.__b
#         c = self.medium.n_ice ** 2 - C_0 ** -2
#         root = np.abs(gamma ** 2 - gamma * b + c)
#         logargument = gamma / (2 * c ** 0.5 * (root) ** 0.5 - b * gamma + 2 * c)
#
#         dydz = 1/(C_0*np.sqrt(c))*(1 - np.sqrt(c)/np.sqrt(root)*(2*gamma-b)*logargument + b*logargument)
#
#         return dydz

    def get_z_from_n(self, n):
        '''
        get z from given n - equation from get_n solved for z
        '''

        return np.log((self.medium.n_ice - n) / self.medium.delta_n) * self.medium.z_0

    def get_surf_skim_angle(self, x1):

        '''
        For a given position x1 = [x,z] and depth profile self.n(), find the angle at which a beam must be
        emitted to "skim the surface", i.e. arrive horizontally (angle = 90 deg) at the surface;
        This is used to find the refraction zone.

        returns:
            C0crit: C0 of critical angle
            thcrit: critical angle
        '''

        nlaunch = self.n(x1[1])
        # by definition, z of critical angle is at surface, i.e. z=0
        zcrit = 0.
        nsurf = self.n(zcrit)

        sinthcrit = nsurf / nlaunch
        if sinthcrit <= 1:
            # ray goes from point with high optical thickness to point with lower optical thickness,
            # i.e. ray bending is towards horizontal
            thcrit = np.arcsin(sinthcrit)
            C0result = self.get_C_0_from_angle(thcrit, x1[1])
            C0crit = C0result.x[0]
        else:
            # ray goes from point with low optical thickness to point with higher optical thickness,
            # i.e. ray bending is towards vertical, no solution
            thcrit = None
            C0crit = None
            self.__logger.warning(' No solution for critical angle for z = {}!'.format(x1[1]))
        self.__logger.info(' critical angle for z = {} is {} !'.format(x1[1], thcrit / units.deg))
        self.__logger.info(' C0 for critical angle is {}'.format(C0crit))

        return C0crit, thcrit

    def is_in_refraction_zone(self, x1, x2, C0crit=None, plot=False):
        '''
        Find if receiver at x2 is in the refraction zone of emitter at x1. The refraction zone
        is the oposite of the shadow zone.

        If the C0 of the critical angle, C0crit, is provided, it will not be calculated. This is useful
        in case find_solutions() is called and C0crit is calculated in the process of determining the
        initial value for the minimization procedure.

        Returns True if x2 is in the refraction zone of x1 - note that the inverse statement is not
        necessarily true, i.e. when False is returned, it is possible that x2 is in the refraction
        zone nonetheless

        TODO:
        Why does the reference point not seem to lie exactly on the mirrored path?
        Instead of returning True/False, it might be useful to return  ycheck - x2[0] (in case x2[0]>ycrit),
        which gives some idea of how close the receiver is to the refraction zone. This could be used to
        define a "gray zone' and a 'far zone', in which the receiver is most definitely in the shadow zone
        '''

        refraction = False

        if C0crit == None:
            C0crit, thcrit = self.get_surf_skim_angle(x1)
        # z_crit = 0 and hence gamma_crit = delta_n by definition
        gcrit = self.medium.delta_n
        # the y-value where the ray hits z=0
        ycrit = self.get_y(gcrit, C0crit, self.get_C_1(x1, C0crit))

        if plot:
            import matplotlib.pyplot as plt
            plt.figure('in_refraction_zone')
            plt.grid(True)
            plt.plot(ycrit, 0, 'ro', label='turning point')
            yarray, zarray = self.get_path(x1, [ycrit, 0], C0crit)
            plt.plot(yarray, zarray, 'ko-', markersize=4, label='path')
            plt.plot(x1[0], x1[1], 'C1d', label='emitter')
            plt.plot(x2[0], x2[1], 'go', label='receiver')

        if x2[0] <= ycrit:
            # not in shadow zone
            refraction = True
            self.__logger.debug(' is_in_refraction_zone(): y-position of receiver smaller than ycrit!')
        else:
            # start horizontal ray at (y,z) = (ycrit,0)
            # experimentally this was found to give slightly different results than mirroring the array at the critical angle.
            # theoretically this is not quite unterstood
            C0check = self.get_C_0_from_angle(np.pi / 2., 0)
            C0check = C0check.x[0]
            gcheck = self.get_gamma(x2[1])
            # print('C0check, gcheck',C0check,gcheck)
            ycheck = -self.get_y(gcheck, C0check, self.get_C_1([ycrit, 0], C0check)) + 2 * ycrit
            # print('ycheck, x2[1]',ycheck,x2[1])
            if x2[0] < ycheck:
                refraction = True
            if plot:
                yarraymirr = -yarray + 2 * ycrit
                plt.plot(yarraymirr, zarray, 'mx-', markersize=4, label='mirrored path')
                # the reference point does not seem to lie exactly on the mirrored path but instead
                # ~1cm inside the path (i.e. towards the emmitter) which I do not understand.
                plt.plot(ycheck, x2[1], 'b+', label='reference point')

        if plot:
            plt.legend(fontsize='x-small')

        return refraction

    def get_tof_for_straight_line(self, x1, x2):
        '''
        Calculate the time of flight for a hypothatical ray travelling from x1 to x2 in a straight line.
        Such an array in general is not a solution consistant with Fermat's principle. It is however
        useful as a reference time or approximation for signals not explicable with geometric optics.

        output:
            time of flight for a ray travelling straight from x1 to x2
        '''

        dx = x2[0] - x1[0]
        dz = x2[1] - x1[1]
        n_ice = self.medium.n_ice
        delta_n = self.medium.delta_n
        z_0 = self.medium.z_0

        if dz > 0:
            return 1. / speed_of_light * np.sqrt((dx / dz) ** 2 + 1) * (
            n_ice * dz - delta_n * z_0 * (np.exp(x2[1] / z_0) - np.exp(x1[1] / z_0))
            )
        else:
            return self.n(x2[1]) / speed_of_light * dx

    def get_surface_pulse(self, x1, x2, infirn=False, angle='critical', chdraw=None, label=None):

        '''
        Calculate the time for a ray to travel from x1 to the surface and arriving at the surface
        with (a) critical angle or (b) Brewster angle, propagating in a straight line along the surface
        in air (n=1) in the firn at z=0 (n=self.n(0)) and then reaching the receiver by returning into the
        firn just at the point to reach the receiver at x2, entering the firn at the surface at the same
        angle it reached the surface from x1..

        Input:
            x1, x2: Arrays with x and z positions of emitter x1 and receiver x2
            infirn: Boolean. Set to True if surface ray travels in the firn, set to False (default) if it travels
                    in air.
            angle:  String specifying angle at which ray reaches/leaves the surface. Can be 'Brewster' or 'critical'
                    If neither of these is chosen, a warning is printed and angle is set to 'critical'
            chdraw: If None, do not draw the path of the ray. If the ray should be drawn, a string consistent with
                    the matplotlib.pyplot library has to be specified, e.g. 'r:' to draw a dotted red line.
                    It is assumed that an appropriate figure on which to draw the ray has been created and set as
                    current figure by the user before calling this method.
            label:  Label for plot
        '''

        draw = False
        if chdraw != None:
            draw = True

        if infirn == False:
            nlayer = 1.  # index of refraction at surface, default is n=1 for air
        else:
            nlayer = self.n(0)

        if angle == 'critical':
            # sin(th)=1,
            nxsin = 1.
        elif angle == 'Brewster':
            nxsin = np.sin(np.arctan(1. / self.n(0))) * self.n(0)
        else:
            self.__logger.warning(' unknown input angle=={}, using critical angle!!!'.format(angle))
            nxsin = 1.

        zsurf = 0
        gamma = self.get_gamma(zsurf)

        # print('nxsin = ',nxsin)
        # find emission angle for starting point x1 to hit the surface at the specified angle

        # look at time and distance it takes for the signal to travel from the emitter to the surface
        # and from the surface to the receiver
        tice = 0
        sice = 0
        for x in [x1, x2]:
            sinthemit = nxsin / self.n(x[1])
            th_emit = np.arcsin(sinthemit)
            C0result = self.get_C_0_from_angle(th_emit, x[1])
            C0_emit = C0result.x[0]

            # print(C0_emit)
            self.__logger.info(' emission angle for position {},{} is theta_emit= {}'.format(x[0], x[1], th_emit / units.deg))

            # x-coordinate where ray reaches surface; is always bigger than the x-position of the emitter
            # (i.e. ray travels "to the right")
            xsurf = self.get_y(gamma, C0_emit, self.get_C_1(x, C0_emit))
            sice += xsurf - x[0]
            self.__logger.info(' air pulse starting at x={}, z={} reaches surface at x={}'.format(x[0], x[1], xsurf))
            ttosurf = self.get_travel_time_analytic(x, [xsurf, zsurf], C0_emit)
            tice += ttosurf
            self.__logger.info(' travel time is {} ns.'.format(ttosurf / units.ns))

            if draw:
                import matplotlib.pyplot as plt
                z = np.linspace(x[1], zsurf, 1000, endpoint=True)
                y = self.get_y(self.get_gamma(z), C0_emit, C_1=self.get_C_1(x, C0_emit))
                if x == x1:
                    ysurf = [y[-1]]
                else:
                    y = -y + 2 * x2[0]
                    ysurf.append(y[-1])
                    plt.plot(ysurf, [0, 0], chdraw, label=label)

                plt.plot(y, z, chdraw)

        self.__logger.info(' time, distance travelled to and from surface: {}, {} '.format(tice, sice))

        sair = abs(x2[0] - x1[0]) - sice
        tair = sair * nlayer / speed_of_light
        self.__logger.info(' time, distance travelled at surface: {}, {}'.format(tair, sair))
        ttot = tice + tair
        if sair < 0:
            ttot = None
        return ttot

    def angular_diff(self, x_refl, z_refl, pulser_pos, receiver_pos, ipulssol, irxsol):

        '''
        This is a helper function to find a ray that is subject to specular reflection (no transmission) at the
        ice-water interface, e.g. for Moore's Bay. For a (virtual) emitter positioned at [x_refl,z_refl], it finds the
        emission angles such that the rays hit positions pulser_pos and receiver_pos. ipulssol = 0 or 1, respectively
        means that pulser_pos is hit directly or by means of reflection at the surface, respectively.
        irxsol is the equivalent parameter for receiver_pos.

        angular_diff can be used as the function to find a root of in the following fashion:

        result = optimize.root(raytr.angular_diff, x0=x_refl_start, args=(z_refl,pulser_pos,receiver_pos,ipulssol,irxsol))

        Then get the final value by x_refl = result.x[0] (z_refl is fixed)

        The idea is to treat the reflection point as a virtual emitter and then find the x-position at the predefined
        depth z_refl, for which the emisssion angles to pulser_pos and receiver_pos are the same (i.e. the output
        of angular_diff is zero). The x-position would be output "result" ofoptimize.root() above.

        output:
               float, is zero if the angles (w.r.t. the vertical) of rays emitted from [x_refl,z_refl] to
               positions pulser_pos and receiver_pos are the same or greater than zero, if this is not the case.
               For exact defintion, see "result" in code below
        '''

        # treat position of reflection as emitter and Rx/Tx as receivers
        pos_rx = [
            [receiver_pos[0], receiver_pos[1]],
            [x_refl - (pulser_pos[0] - x_refl), pulser_pos[1]]
        ]
        beta0 = None
        beta1 = None
        # solution for receiver
        solution0 = self.find_solutions([x_refl, z_refl], pos_rx[0], plot=False)
        if solution0 != []:
            C0rx = solution0[irxsol]['C0']
            beta0 = self.get_launch_angle([x_refl, z_refl], C0rx)
        # solution for pulser
        solution1 = self.find_solutions([x_refl, z_refl], pos_rx[1], plot=False)
        if solution1 != []:
            C0puls = solution1[ipulssol]['C0']
            beta1 = self.get_launch_angle([x_refl, z_refl], C0puls)

        if beta0 != None and beta1 != None:
            result = (np.tan(beta0) - np.tan(beta1)) ** 2
        else:
            result = np.inf  # set to infinity

        return result


class ray_tracing:
    """
    utility class (wrapper around the 2D analytic ray tracing code) to get
    ray tracing solutions in 3D for two arbitrary points x1 and x2
    """
    solution_types = {1: 'direct',
                      2: 'refracted',
                      3: 'reflected'}

    def __init__(self, x1, x2, medium, attenuation_model="SP1", log_level=logging.WARNING,
                 n_frequencies_integration=6,
                 n_reflections=0):
        """
        class initilization

        Parameters
        ----------
        x1: 3dim np.array
            start point of the ray
        x2: 3dim np.array
            stop point of the ray
        medium: medium class
            class describing the index-of-refraction profile
        attenuation_model: string
            signal attenuation model (so far only "SP1" is implemented)
        log_level: logging object
            specify the log level of the ray tracing class
            * logging.ERROR
            * logging.WARNING
            * logging.INFO
            * logging.DEBUG
            default is WARNING
        n_frequencies_integration: int
            the number of frequencies for which the frequency dependent attenuation
            length is being calculated. The attenuation length for all other frequencies
            is obtained via linear interpolation.
            
        n_reflections: int (default 0)
            in case of a medium with a reflective layer at the bottom, how many reflections should be considered

        """
        self.__logger = logging.getLogger('ray_tracing')
        self.__logger.setLevel(log_level)
        self.__medium = medium
        self.__attenuation_model = attenuation_model
        self.__n_frequencies_integration = n_frequencies_integration
        if(n_reflections):
            if(not hasattr(self.__medium, "reflection") or self.__medium.reflection is None):
                self.__logger.warning("ray paths with bottom reflections requested medium does not have any reflective layer, setting number of reflections to zero.")
                n_reflections = 0
        self.__n_reflections = n_reflections
        if(n_reflections):
            if(x1[2] < self.__medium.reflection  or x2[2] < self.__medium.reflection):
                self.__logger.error("start or stop point is below the reflective layer at {:.1f}m".format(self.__medium.reflection / units.m))
                raise AttributeError("start or stop point is below the reflective layer at {:.1f}m".format(self.__medium.reflection / units.m))

        self.__swap = False
        self.__X1 = x1
        self.__X2 = x2
        if(x2[2] < x1[2]):
            self.__swap = True
            self.__logger.debug('swap = True')
            self.__X2 = x1
            self.__X1 = x2

        dX = self.__X2 - self.__X1
        self.__dPhi = -np.arctan2(dX[1], dX[0])
        c, s = np.cos(self.__dPhi), np.sin(self.__dPhi)
        self.__R = np.array(((c, -s, 0), (s, c, 0), (0, 0, 1)))
        X1r = self.__X1
        X2r = np.dot(self.__R, self.__X2 - self.__X1) + self.__X1
        self.__logger.debug("X1 = {}, X2 = {}".format(self.__X1, self.__X2))
        self.__logger.debug('dphi = {:.1f}'.format(self.__dPhi / units.deg))
        self.__logger.debug("X2 - X1 = {}, X1r = {}, X2r = {}".format(self.__X2 - self.__X1, X1r, X2r))
        self.__x1 = np.array([X1r[0], X1r[2]])
        self.__x2 = np.array([X2r[0], X2r[2]])

        self.__logger.debug("2D points {} {}".format(self.__x1, self.__x2))
        self.__r2d = ray_tracing_2D(self.__medium, self.__attenuation_model, log_level=log_level,
                                    n_frequencies_integration=self.__n_frequencies_integration)

    def set_solution(self, C0s, C1s, solution_types, reflection=None, reflection_case=None):
        results = []
        if(reflection is None):
            reflection = np.zeros_like(C0s, dtype=np.int)
            reflection_case = np.ones_like(C0s, dtype=np.int)
        for i in range(len(C0s)):
            if(not np.isnan(C0s[i])):
                results.append({'type': solution_types[i],
                                'C0': C0s[i],
                                'C1': C1s[i],
                                'reflection': reflection[i],
                                'reflection_case': reflection_case[i]})
        self.__results = results

    def find_solutions(self):
        """
        find all solutions between x1 and x2
        """
        self.__results = self.__r2d.find_solutions(self.__x1, self.__x2)
        for i in range(self.__n_reflections):
            for j in range(2):
                self.__results.extend(self.__r2d.find_solutions(self.__x1, self.__x2, reflection=i + 1, reflection_case=j + 1))

    def has_solution(self):
        """
        checks if ray tracing solution exists
        """
        return len(self.__results) > 0

    def get_number_of_solutions(self):
        """
        returns the number of solutions
        """
        return len(self.__results)

    def get_results(self):
        """
        returns dictionary of results (the parameters of the analytic ray path function)
        """
        return self.__results

    def get_solution_type(self, iS):
        """ returns the type of the solution

        Parameters
        ----------
        iS: int
            choose for which solution to compute the launch vector, counting
            starts at zero

        Returns
        -------
        solution_type: int
            * 1: 'direct'
            * 2: 'refracted'
            * 3: 'reflected
        """
        return self.__r2d.determine_solution_type(self.__x1, self.__x2, self.__results[iS]['C0'])

    def get_path(self, iS, n_points=1000):
        n = self.get_number_of_solutions()
        if(iS >= n):
            self.__logger.error("solution number {:d} requested but only {:d} solutions exist".format(iS + 1, n))
            raise IndexError
        result = self.__results[iS]
        xx, zz = self.__r2d.get_path_reflections(self.__x1, self.__x2, result['C0'], n_points=n_points,
                                                 reflection=result['reflection'],
                                                 reflection_case=result['reflection_case'])
        path_2d = np.array([xx, np.zeros_like(xx), zz]).T
        dP = path_2d - np.array([self.__X1[0], 0, self.__X1[2]])
        MM = np.matmul(self.__R.T, dP.T)
        path = MM.T + self.__X1
        return path

    def get_launch_vector(self, iS):
        """
        calculates the launch vector (in 3D) of solution iS

        Parameters
        ----------
        iS: int
            choose for which solution to compute the launch vector, counting
            starts at zero

        Returns
        -------
        launch_vector: 3dim np.array
            the launch vector
        """
        n = self.get_number_of_solutions()
        if(iS >= n):
            self.__logger.error("solution number {:d} requested but only {:d} solutions exist".format(iS + 1, n))
            raise IndexError

        result = self.__results[iS]
        alpha = self.__r2d.get_launch_angle(self.__x1, result['C0'], reflection=result['reflection'],
                                            reflection_case=result['reflection_case'])
        launch_vector_2d = np.array([np.sin(alpha), 0, np.cos(alpha)])
        if self.__swap:
            alpha = self.__r2d.get_receive_angle(self.__x1, self.__x2, result['C0'],
                                                 reflection=result['reflection'],
                                                 reflection_case=result['reflection_case'])
            launch_vector_2d = np.array([-np.sin(alpha), 0, np.cos(alpha)])
        self.__logger.debug(self.__R.T)
        launch_vector = np.dot(self.__R.T, launch_vector_2d)
        return launch_vector

    def get_receive_vector(self, iS):
        """
        calculates the receive vector (in 3D) of solution iS

        Parameters
        ----------
        iS: int
            choose for which solution to compute the launch vector, counting
            starts at zero

        Returns
        -------
        receive_vector: 3dim np.array
            the receive vector
        """
        n = self.get_number_of_solutions()
        if(iS >= n):
            self.__logger.error("solution number {:d} requested but only {:d} solutions exist".format(iS + 1, n))
            raise IndexError

        result = self.__results[iS]
        alpha = self.__r2d.get_receive_angle(self.__x1, self.__x2, result['C0'],
                                             reflection=result['reflection'],
                                             reflection_case=result['reflection_case'])
        receive_vector_2d = np.array([-np.sin(alpha), 0, np.cos(alpha)])
        if self.__swap:
            alpha = self.__r2d.get_launch_angle(self.__x1, result['C0'],
                                                reflection=result['reflection'],
                                                reflection_case=result['reflection_case'])
            receive_vector_2d = np.array([np.sin(alpha), 0, np.cos(alpha)])
        receive_vector = np.dot(self.__R.T, receive_vector_2d)
        return receive_vector

    def get_reflection_angle(self, iS):
        """
        calculates the angle of reflection at the surface (in case of a reflected ray)

        Parameters
        ----------
        iS: int
            choose for which solution to compute the launch vector, counting
            starts at zero

        Returns
        -------
        reflection_angle: float or None
            the reflection angle (for reflected rays) or None for direct and refracted rays
        """
        n = self.get_number_of_solutions()
        if(iS >= n):
            self.__logger.error("solution number {:d} requested but only {:d} solutions exist".format(iS + 1, n))
            raise IndexError

        result = self.__results[iS]
        return self.__r2d.get_reflection_angle(self.__x1, self.__x2, result['C0'],
                                               reflection=result['reflection'], reflection_case=result['reflection_case'])

    def get_path_length(self, iS, analytic=True):
        """
        calculates the path length of solution iS

        Parameters
        ----------
        iS: int
            choose for which solution to compute the launch vector, counting
            starts at zero

        analytic: bool
            If True the analytic solution is used. If False, a numerical integration is used. (default: True)

        Returns
        -------
        distance: float
            distance from x1 to x2 along the ray path
        """
        n = self.get_number_of_solutions()
        if(iS >= n):
            self.__logger.error("solution number {:d} requested but only {:d} solutions exist".format(iS + 1, n))
            raise IndexError

        result = self.__results[iS]
        if analytic:
            analytic_length = self.__r2d.get_path_length_analytic(self.__x1, self.__x2, result['C0'],
                                                                  reflection=result['reflection'],
                                                                  reflection_case=result['reflection_case'])
            if (analytic_length != None):
                return analytic_length
            else:
                return self.__r2d.get_path_length(self.__x1, self.__x2, result['C0'],
                                                  reflection=result['reflection'],
                                                  reflection_case=result['reflection_case'])
        else:
            return self.__r2d.get_path_length(self.__x1, self.__x2, result['C0'],
                                              reflection=result['reflection'],
                                              reflection_case=result['reflection_case'])

    def get_travel_time(self, iS, analytic=True):
        """
        calculates the travel time of solution iS

        Parameters
        ----------
        iS: int
            choose for which solution to compute the launch vector, counting
            starts at zero

        analytic: bool
            If True the analytic solution is used. If False, a numerical integration is used. (default: True)

        Returns
        -------
        time: float
            travel time
        """
        n = self.get_number_of_solutions()
        if(iS >= n):
            self.__logger.error("solution number {:d} requested but only {:d} solutions exist".format(iS + 1, n))
            raise IndexError

        result = self.__results[iS]
        if(analytic):
            analytic_time = self.__r2d.get_travel_time_analytic(self.__x1, self.__x2, result['C0'],
                                                                reflection=result['reflection'],
                                                                reflection_case=result['reflection_case'])
            if (analytic_time != None):
                return analytic_time
            else:
                return self.__r2d.get_travel_time(self.__x1, self.__x2, result['C0'],
                                                  reflection=result['reflection'],
                                                  reflection_case=result['reflection_case'])
        else:
            return self.__r2d.get_travel_time(self.__x1, self.__x2, result['C0'],
                                              reflection=result['reflection'],
                                              reflection_case=result['reflection_case'])

    def get_attenuation(self, iS, frequency, max_detector_freq=None):
        """
        calculates the signal attenuation due to attenuation in the medium (ice)

        Parameters
        ----------
        iS: int
            choose for which solution to compute the launch vector, counting
            starts at zero

        frequency: array of floats
            the frequencies for which the attenuation is calculated

        max_detector_freq: float or None
            the maximum frequency of the final detector sampling
            (the simulation is internally run with a higher sampling rate, but the relevant part of the attenuation length
            calculation is the frequency interval visible by the detector, hence a finer calculation is more important)

        Returns
        -------
        attenuation: array of floats
            the fraction of the signal that reaches the observer
            (only ice attenuation, the 1/R signal falloff not considered here)
        """
        n = self.get_number_of_solutions()
        if(iS >= n):
            self.__logger.error("solution number {:d} requested but only {:d} solutions exist".format(iS + 1, n))
            raise IndexError

        result = self.__results[iS]
        return self.__r2d.get_attenuation_along_path(self.__x1, self.__x2, result['C0'], frequency, max_detector_freq,
                                                     reflection=result['reflection'],
                                                     reflection_case=result['reflection_case'])

    def get_ray_path(self, iS):
        return self.__r2d.get_path(self.__x1, self.__x2, self.__results[iS]['C0'], 10000,
                                   reflection=self.__results[iS]['reflection'],
                                   reflection_case=self.__results[iS]['reflection_case'])



#!/usr/bin/env python
import numpy as np
import h5py
from NuRadioMC.utilities import units
import sys
import os
path = os.path.dirname(os.path.abspath(__file__))

fin = h5py.File(os.path.join(path,"output.hdf5"), 'r')
weights = np.array(fin['weights'])
n_events = fin.attrs['n_events']

###########################
# calculate effective volume
###########################
density_ice = 0.9167 * units.g / units.cm ** 3
density_water = 997 * units.kg / units.m ** 3

n_triggered = np.sum(weights)

V = None
rmin = fin.attrs['rmin']
rmax = fin.attrs['rmax']
dZ = fin.attrs['zmax'] - fin.attrs['zmin']
V = np.pi * (rmax ** 2 - rmin ** 2) * dZ
Veff = V * density_ice / density_water * 4 * np.pi * np.sum(weights) / n_events
print('fraction of triggered events = {:.0f}/{:.0f} = {:.3f} -> {:.6g} km^3 sr'.format(n_triggered, n_events, n_triggered / n_events, Veff / units.km ** 3))

Veff_mean = 4.74
Veff_sigma = 0.43

delta = (Veff / units.km ** 3 - Veff_mean) / Veff_sigma
rdelta = (Veff / units.km ** 3 - Veff_mean) / Veff_mean
print("effective volume deviates {:.1f} sigma ({:.0f}%) from the mean".format(delta, 100 * rdelta))

if(np.abs(Veff / units.km ** 3 - Veff_mean) > 2 * Veff_sigma):
    print("deviation is more than 2 sigma -> this should only happen in 5\% of the tests. Rerun the test and see if the error persists.")
    sys.exit(-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
from NuRadioReco.utilities import units
from scipy import interpolate as intp
from scipy import integrate as int
from scipy import constants
from matplotlib import pyplot as plt
import os
import pickle
from time import time
from NuRadioMC.SignalGen.ARZ.ARZ import *

energy = 1.e6 * units.TeV
theta = 55 * units.deg
R = 1 * units.km
N = 512
dt = 0.1 * units.ns
n_index = 1.78
y = 0
ccnc = 'cc'
flavor = 12  # e = 12, mu = 14, tau = 16

cdir = os.path.dirname(__file__)
bins, depth_e, N_e = np.loadtxt(os.path.join(cdir, "../shower_library/nue_1EeV_CC_1_s0001.t1005"), unpack=True)
bins, depth_p, N_p = np.loadtxt(os.path.join(cdir, "../shower_library/nue_1EeV_CC_1_s0001.t1006"), unpack=True)
depth_e *= units.g / units.cm**2
depth_p *= units.g / units.cm**2
depth_e -= 1000 * units.g/units.cm**2  # all simulations have an artificial offset of 1000 g/cm^2
depth_p -= 1000 * units.g/units.cm**2
# sanity check if files electron and positron profiles are compatible
if (not np.all(depth_e == depth_p)):
    raise ImportError("electron and positron profile have different depths")

t0 = time()
vp2 = get_vector_potential_fast(energy, theta, N, dt, "EM", n_index, R, profile_depth=depth_e, profile_ce=(N_e-N_p))
print("fast calculation took {:.1f} ms".format((time() -t0)*1e3))
t0 = time()
vp = get_vector_potential(energy, theta, N, dt, y, ccnc, flavor, n_index, R, profile_depth=depth_e, profile_ce=(N_e-N_p))
print("slow calculation took {:.4f}s".format(time() -t0))

# generate time array
tt = np.arange(0, (N + 1) * dt, dt)
tt = tt + 0.5 * dt - tt.mean()

fig, ax = plt.subplots(1, 1)
ax.plot(tt, vp[:, 0] / units.V / units.s)
ax.plot(tt, vp[:, 1] / units.V / units.s)
ax.plot(tt, vp[:, 2] / units.V / units.s)
ax.plot(tt, vp2[:, 0] / units.V / units.s, "C0--")
ax.plot(tt, vp2[:, 1] / units.V / units.s, "C1--")
ax.plot(tt, vp2[:, 2] / units.V / units.s, "C2--")
ax.set_xlim(-2, 2)

ax.set_xlabel("time [ns]")
ax.set_ylabel("vector potential")


fig, ax = plt.subplots(1, 1)
ax.plot(tt, vp[:, 0]/vp2[:, 0])
ax.plot(tt, vp[:, 2]/ vp2[:, 0])
ax.set_xlim(-2, 2)
ax.set_xlabel("time [ns]")
ax.set_ylabel("python/fortran implementation")
ax.set_ylim(0.8, 1.2)
plt.show()


import NuRadioMC.SignalGen.RalstonBuniy.askaryan_module as AskaryanModule
import NuRadioMC.SignalGen.parametrizations as param
from NuRadioMC.utilities import units, fft
import numpy as np
import matplotlib.pyplot as plt
from radiotools import plthelpers as php

model = 'Alvarez2000'

dt = .1 * units.ns
n_samples = 2048
ff = np.fft.rfftfreq(n_samples, dt)
tt = np.arange(0, n_samples * dt, dt)
df = ff[1]  # width of a frequency bin

if(model == 'Alvarez2000'):
    from NuRadioMC.SignalGen.RalstonBuniy.askaryan_module import get_frequency_spectrum

# plot dependence on cherenkov cone
E = 1e18 * units.eV
n_index = 1.78
R = 1 * units.km
domegas = np.array([0, 1, 2, 3, 4, 5, 7.5, 10]) * units.deg
emstr = ['EM', 'had.']
em = np.array([True, False])
# lpm = np.array([False, True, False])
for i in range(2):  # loop over EM/EM+LPM/HAD
    fig, ax = plt.subplots(1, 1)
    for iOmega, domega in enumerate(domegas):
        theta = np.arccos(1. / n_index) + domega
        spec1 = param.get_frequency_spectrum(E, theta, n_samples, dt, em[i], n_index, R, model=model)
        ax.plot(ff / units.MHz, np.abs(spec1) / units.V * units.m, php.get_color_linestyle(iOmega), label='$\Delta \Omega$ = {:.1f}deg'.format(domega / units.deg))
    ax.semilogx(True)
    ax.semilogy(True)
    ax.set_xlim(10, 2e3)
    ax.set_ylim(1e-7, 1e-1)
    ax.set_xlabel("frequency [MHz]")
    ax.set_ylabel("amplitude [V/m] per {:.1f}MHz".format(df / units.MHz))
    ax.legend(fontsize='small')
    ax.set_title("{} E = {:.1e}eV, R = {:.1f}km, {}".format(model, E / units.eV, R / units.km, emstr[i]))
    fig.tight_layout()
    fig.savefig("plots/{}_E{:.1e}eV_R{:.0f}m_EM{}.png".format(model, E / units.eV, R / units.m, em[i]))
    plt.show()
    plt.close('all')

# plot distance dependence for domegas
E = 1e16 * units.eV
n_index = 1.78
domegas = np.array([0, 2.5, 5]) * units.deg
em = np.array([True, True, False])
lpm = np.array([False, True, False])
Rs = np.array([100, 200, 500, 1000, 2000]) * units.m
for i in range(3):  # loop over EM/EM+LPM/HAD
    for iOmega, domega in enumerate(domegas):
        theta = np.arccos(1. / n_index) + domega
        fig, ax = plt.subplots(1, 1)
        for iR, R in enumerate(Rs):
            spec1 = get_frequency_spectrum(E, theta, ff, em[i], n_index, R, lpm[i]) * R
            ax.plot(ff / units.MHz, np.abs(spec1[1]) / units.V * units.m, php.get_color_linestyle(iR), label='R = {:.0f}m'.format(R / units.m))
        ax.semilogx(True)
        ax.semilogy(True)
        ax.set_xlim(10, 2e3)
        ax.set_ylim(1e-6, 1e-0)
        ax.set_xlabel("frequency [MHz]")
        ax.set_ylabel("amplitude [V/m] * R [m] per {:.1f}MHz".format(df / units.MHz))
        ax.legend(fontsize='small')
        ax.set_title("{} E = {:.1e}eV, $\Delta \Omega$ = {:.1f}deg, {}".format(model, E / units.eV, domega / units.deg, emstr[i]))
        fig.tight_layout()
        fig.savefig("plots/{}_E{:.1e}eV_dOmega{:.1f}_EM{}_LPM{}.png".format(model, E / units.eV, domega / units.deg, em[i], lpm[i]))
        plt.close('all')



from __future__ import absolute_import, division, print_function
import numpy as np
from matplotlib import pyplot as plt
from NuRadioMC.utilities import units

def plot_vertex_distribution(xx, yy, zz, weights=None,
                            rmax=None, zmin=None,
                            trigger_name=""):
    """
    produces beautiful plot of the vertex distribution

    Parameters:
    xx: array
        the x positions of the vertices
    yy: array
        the y positions of the vertices
    zz: array
        the z positions of the vertices

    Returns
    -------
    fig, ax
    """
    if(weights is None):
        weights = np.ones_like(xx)
    fig, ax = plt.subplots(1, 1)
    rr = (xx ** 2 + yy ** 2) ** 0.5
    mask_weight = weights > 1e-2
    max_r = rr[mask_weight].max()
    max_z = np.abs(zz[mask_weight]).max()
    if(rmax is None):
        rmax = max_r
    if(zmin is None):
        zmin = zz.min()
    h = ax.hist2d(rr / units.m, zz / units.m, bins=[np.linspace(0, max_r, 50), np.linspace(-max_z, 0, 50)],
                  cmap=plt.get_cmap('Blues'), weights=weights)
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cb = plt.colorbar(h[3], ax=ax, cax=cax)
    cb.set_label("weighted number of events")
    ax.set_aspect('equal')
    ax.set_xlabel("r [m]")
    ax.set_ylabel("z [m]")
    ax.set_xlim(0, rmax)
    ax.set_ylim(zmin, 0)
    if(trigger_name != ""):
        ax.set_title("trigger: {}".format(trigger_name))
    fig.tight_layout()
    return fig, ax



from __future__ import absolute_import, division, print_function
from NuRadioMC.utilities import units
from NuRadioMC.EvtGen.generator import generate_eventlist_cylinder

# define simulation volume
zmin = -2.7 * units.km  # the ice sheet at South Pole is 2.7km deep
zmax = 0 * units.km
rmin = 0 * units.km
rmax = 4 * units.km

# generate one event list at 1e19 eV with 1000 neutrinos
generate_eventlist_cylinder('1e19_n1e3.hdf5', 1e3, 1e19 * units.eV, 1e19 * units.eV, rmin, rmax, zmin, zmax,
                             full_rmin=rmin, full_rmax=rmax, full_zmin=zmin, full_zmax=zmax)

# generate one event list at 1e18 eV with 10000 neutrinos
generate_eventlist_cylinder('1e18_n1e4.hdf5', 1e4, 1e18 * units.eV, 1e18 * units.eV, rmin, rmax, zmin, zmax,
                            full_rmin=rmin, full_rmax=rmax, full_zmin=zmin, full_zmax=zmax)


from NuRadioMC.utilities import units
import matplotlib.pyplot as plt
import numpy as np
import json

"""
This file plots the tau energy and length profiles percentiles from the
file calculated with PROPOSAL using T08_tau_PROPOSAL.py and also the profiles
from the CSDA calculated using T03_tau_length.py. It also performs a fit to the
tau length mean.

Just run it by typing:
    python T09_plot_PROPOSAL_diff.py
"""

def get_dict(filename):

    with open(filename, 'r') as datafile:
        input = json.load(datafile)

    for key in input:
        input[key] = np.array(input[key])

    return input

file_cont = 'continuous_decay.json'
file_tau = 'tau_proposal_decay.json'
file_taubar = 'taubar_proposal_decay.json'

input_cont = get_dict(file_cont)
input_tau = get_dict(file_tau)
#input_taubar = get_dict(file_taubar)

# PLOTTING ENERGIES

plt.tight_layout()

plt.gcf().subplots_adjust(bottom=0.13)
plt.tick_params(labelsize=12)

plt.loglog(input_cont['energies'], input_cont['tau_energies_mean'], linestyle='-', color='black', label='Mean with cont. PN losses')
plt.loglog(input_tau['energies'], input_tau['tau_energies_mean'], linestyle='-', color='red', label='Mean PROPOSAL with all losses')
plt.loglog(input_cont['energies'], input_cont['energies'], linestyle='-.', color='red', label='Mean, without losses')

plt.fill_between(input_cont['energies'], input_cont['tau_energies_10'], input_cont['tau_energies_90'],
                facecolor='rosybrown', alpha=0.75, interpolate=True, label=r'10% to 90% quantiles, PN')
plt.fill_between(input_tau['energies'], input_tau['tau_energies_10'], input_tau['tau_energies_90'],
                facecolor='skyblue', alpha=0.75, interpolate=True, label=r'10% to 90% quantiles, PROPOSAL')

plt.xlabel('Tau initial energy [eV]', size=16)
plt.ylabel('Tau decay energy [eV]', size=16)
plt.legend(fontsize=11,loc=2)
plt.ylim((1e14, None))
plt.savefig('tau_decay_energy_PROPOSAL_diff.png', format='png')
plt.show()

# PLOTTING LENGTHS

from scipy import constants
tau_rest_lifetime = 290.3 * units.fs
cspeed = constants.c * units.m / units.s
tau_mass = constants.physical_constants['tau mass energy equivalent in MeV'][0] * units.MeV
length_no_losses = input_cont['energies']*tau_rest_lifetime*cspeed / tau_mass

plt.tight_layout()

plt.gcf().subplots_adjust(bottom=0.13)
plt.tick_params(labelsize=12)

plt.loglog(input_cont['energies'], input_cont['lengths_mean']/units.km, linestyle='-', color='black', label='Mean with cont. PN losses')
plt.loglog(input_tau['energies'], input_tau['lengths_mean']/units.km, linestyle='-', color='red', label='Mean PROPOSAL with all losses')
plt.loglog(input_cont['energies'], length_no_losses/units.km, linestyle='-.', color='red', label='Mean, without losses')

plt.fill_between(input_cont['energies'], input_cont['lengths_10']/units.km, input_cont['lengths_90']/units.km,
                facecolor='rosybrown', alpha=0.75, interpolate=True, label=r'10% to 90% quantiles, PN')
plt.fill_between(input_tau['energies'], input_tau['lengths_10']/units.km, input_tau['lengths_90']/units.km,
                facecolor='skyblue', alpha=0.75, interpolate=True, label=r'10% to 90% quantiles, PROPOSAL')

plt.xlabel('Tau initial energy [eV]', size=16)
plt.ylabel('Tau track length [km]', size=16)
plt.legend(fontsize=12)
plt.savefig('tau_decay_length_PROPOSAL_diff.png', format='png')
plt.show()

# FITTING THE 95% PERCENTILE FOR TAU SIMULATIONS

from scipy.optimize import curve_fit

def poly(x, a, b, c, d, e):
    return a + b*x + c*x**2 + d*x**3 + e*x**4

xdata = np.log10(input_tau['energies'])
ydata = np.log10(input_tau['lengths_95'])

results = curve_fit(poly, xdata, ydata)[0]
print("Results for the polynomial fit", results)

plt.plot(input_tau['energies'], input_tau['lengths_95']/units.km, label='95% quantile by PROPOSAL')
plt.plot(input_tau['energies'], 10**poly(xdata, *results)/units.km, label='Polynomial fit')
plt.xscale('log')
plt.yscale('log')
plt.xlabel('Tau initial energy [eV]', size=16)
plt.ylabel('Tau track length [km]', size=16)
plt.legend()
plt.show()



import NuRadioMC.EvtGen.NuRadioProposal as NRP
from NuRadioMC.utilities import units
import matplotlib.pyplot as plt
import numpy as np
import json

"""
This file calculates some percentiles for the tau length and energy distributions
given by PROPOSAL, plots them, and saves them to a json file.

Just run it by typing:
    python T08_tau_PROPOSAL.py
The number of simulated muons per energy can be changed with Ntries, and the number
of energies can be changed with Nenergies
"""

Ntries = 1000
Nenergies = 20

lepton_code = 15
filename = 'tau_proposal_decay.json'

energies = np.linspace(15,20,Nenergies)
energies = 10**energies * units.eV

dists_arr = []
energy_arr = []

for Elepton in energies:

    decay_prods = NRP.GetDecays([Elepton]*Ntries, [lepton_code]*Ntries, random_seed=None)

    dist_hist = decay_prods[:,0]
    E_hist = decay_prods[:,1]

    # Filter energies below 0.1 PeV
    mask = E_hist < 0.1*units.PeV
    E_hist[mask] = 0.1*units.PeV

    dists_arr.append(dist_hist)
    energy_arr.append(E_hist)

    print("energy: ", Elepton/units.PeV)

dists_arr = np.array(dists_arr)
energies_arr = np.array(energy_arr)

print(dists_arr.shape)
print(energies_arr.shape)

dists_05 = np.quantile(dists_arr, 0.05, axis=1)
dists_10 = np.quantile(dists_arr, 0.1, axis=1)
dists_16 = np.quantile(dists_arr, 0.16, axis=1)
dists_mean = np.mean(dists_arr, axis=1)
dists_84 = np.quantile(dists_arr, 0.84, axis=1)
dists_90 = np.quantile(dists_arr, 0.9, axis=1)
dists_95 = np.quantile(dists_arr, 0.95, axis=1)

energies_05 = np.quantile(energies_arr, 0.05, axis=1)
energies_10 = np.quantile(energies_arr, 0.1, axis=1)
energies_16 = np.quantile(energies_arr, 0.16, axis=1)
energies_mean = np.mean(energies_arr, axis=1)
energies_84 = np.quantile(energies_arr, 0.84, axis=1)
energies_90 = np.quantile(energies_arr, 0.9, axis=1)
energies_95 = np.quantile(energies_arr, 0.95, axis=1)

plot = False
if plot:
    plt.tight_layout()

    plt.gcf().subplots_adjust(bottom=0.13)
    plt.tick_params(labelsize=12)

    plt.loglog(energies, energies_mean, linestyle='-', color='red', label='Mean PROPOSAL with all losses')
    plt.loglog(energies, energies, linestyle='--', color='red', label='Mean, without losses')

    plt.fill_between(energies, energies_10, energies_90, facecolor='0.75', interpolate=True, label=r'10% to 90% quantiles')

    plt.xlabel('Tau initial energy [eV]', size=16)
    plt.ylabel('Tau decay energy [eV]', size=16)
    plt.legend(fontsize=12)
    plt.savefig('tau_decay_energy_PROPOSAL.png', format='png')
    plt.show()

output = {}
output['energies'] = list(energies)
output['lengths_05'] = list(dists_05)
output['lengths_10'] = list(dists_10)
output['lengths_16'] = list(dists_16)
output['lengths_84'] = list(dists_84)
output['lengths_90'] = list(dists_90)
output['lengths_95'] = list(dists_95)
output['lengths_mean'] = list(dists_mean)
output['tau_energies_05'] = list(energies_05)
output['tau_energies_10'] = list(energies_10)
output['tau_energies_16'] = list(energies_16)
output['tau_energies_84'] = list(energies_84)
output['tau_energies_90'] = list(energies_90)
output['tau_energies_95'] = list(energies_95)
output['tau_energies_mean'] = list(energies_mean)

with open(filename, 'w+') as data_file:
    json.dump(output, data_file, sort_keys=True, indent=4)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['NuRadioMC',
 'NuRadioMC.EvtGen',
 'NuRadioMC.SignalGen',
 'NuRadioMC.SignalGen.ARZ',
 'NuRadioMC.SignalProp',
 'NuRadioMC.SignalProp.CPPAnalyticRayTracing',
 'NuRadioMC.examples',
 'NuRadioMC.examples.02_DnR',
 'NuRadioMC.examples.Sensitivities',
 'NuRadioMC.simulation',
 'NuRadioMC.studies',
 'NuRadioMC.utilities']

package_data = \
{'': ['*'],
 'NuRadioMC': ['SignalGenProp/*',
               'data/*',
               'test/*',
               'test/SignalGen/*',
               'test/SignalProp/*',
               'test/SingleEvents/*',
               'test/Veff/*',
               'test/Veff/1e18eV/*',
               'test/examples/*'],
 'NuRadioMC.EvtGen': ['PROPOSAL_tables/*', 'tests/*'],
 'NuRadioMC.SignalGen': ['tests/*', 'tests/plots/*'],
 'NuRadioMC.SignalGen.ARZ': ['shower_library/*', 'tests/*'],
 'NuRadioMC.SignalProp': ['examples/*'],
 'NuRadioMC.examples': ['01_Veff_simulation/*',
                        '03_station_coincidences/*',
                        '04_read_ari_files/*',
                        '05_pulser_calibration_measurement/*',
                        'cluster/*'],
 'NuRadioMC.examples.02_DnR': ['detector/*', 'event_input/*'],
 'NuRadioMC.examples.Sensitivities': ['RNO_Proposal/*'],
 'NuRadioMC.simulation': ['detectors/*',
                          'detectors/ARA/*',
                          'detectors/depth/*']}

install_requires = \
['radiotools >=0.2.0', 'h5py', 'NuRadioReco >=1.0.0', 'pyyaml', 'cython']

setup(name='NuRadioMC',
      version='1.0.0',
      description='NuRadioMC: Simulating the radio emission of neutrinos from interaction to detector',
      author='Christian Glaser',
      author_email='work@c-glaser.de',
      url='https://github.com/nu-radio/NuRadioMC',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>3.4',
     )



from __future__ import print_function, division
import numpy as np
from NuRadioMC.utilities import units
from NuRadioReco.utilities import io_utilities
import os
from scipy import interpolate as intp
import glob
import pickle
import sys
from radiotools import plthelpers as php
from matplotlib import pyplot as plt
from NuRadioMC.SignalGen.ARZ import ARZ
import logging
from multiprocessing import Process, Pool
from scipy import signal
import matplotlib.gridspec as gridspec

logger = logging.getLogger("test")
logging.basicConfig(level=logging.WARNING)
rho = 0.924 * units.g / units.cm ** 3  # density g cm^-3

sampling_rate = 5 * units.GHz
dt = 1. / sampling_rate
T2 = 50 * units.ns
N2 = np.int(np.round(T2 / dt))
n_index = 1.78
theta = np.arccos(1. / n_index)
R = 5 * units.km
T = 600 * units.ns
N = np.int(np.round(T / dt))
tt = np.arange(0, dt * N, dt)

dCs = np.append(np.arange(-20, -5, 1), np.append(np.arange(-5,-1,0.2), np.arange(-1, 0, .05)))
dCs = np.append(np.append(dCs, [0]), -1 * dCs)
dCs = np.sort(dCs) * units.deg

plot=False

def calculate(shower_type, E, nE):
    lib_path = "/Users/cglaser/work/ARIANNA/data/ARZ/v1.1/library_v1.1.pkl"
    a = ARZ.ARZ(library=lib_path, interp_factor=100)
    showers['meta'] = {'dt': dt,
                       'n_index': n_index,
                       'R': R}
    showers[shower_type] = {}
    if(E not in showers[shower_type]):
        showers[shower_type][E] = {}
    for iN in range(nE):  # loop through shower realizations
        print("{} {:.2g} {:d}".format(shower_type, E, iN))
        if(iN not in showers[shower_type][E]):
            showers[shower_type][E][iN] = {}
        if plot:
            fig, ax = plt.subplots(1, 1)
            fig2, ax2 = plt.subplots(1, 1)
        for dC in dCs:  # loop through different cherenkov angles
            theta1 = theta + dC
            trace, Lmax = a.get_time_trace(E, theta1, N, dt, shower_type, n_index, R, iN=iN, output_mode='Xmax')
            iMax = np.argmax(np.abs(trace[1]))
            i1 = iMax - N2 // 2
            if((i1 + N2) >= N) or (i1 < 0):
                print("tracelength to short, i1 = {}".format(i1))
#                             raise IndexError("tracelength to short, i1 = {}".format(i1))
            t0 = tt[i1] - tt.mean()
            showers[shower_type][E][iN][theta1] = {}
            showers[shower_type][E][iN][theta1]['t0'] = t0
            showers[shower_type][E][iN][theta1]['Lmax'] = Lmax
            showers[shower_type][E][iN][theta1]['trace'] = trace[1][i1:(i1 + N2)]
            if(plot):
                ax.plot(tt[i1:(i1 + N2)], trace[1][i1:(i1 + N2)] / np.abs(trace[1]).max())
                ax2.plot(tt[i1:(i1 + N2)], np.abs(signal.hilbert(trace[1][i1:(i1 + N2)])))
        if(plot):
            ax.set_title("{} E = {:.2g}eV i = {:d}".format(shower_type, E / units.eV, iN))
            ax2.set_title("{} E = {:.2g}eV i = {:d}".format(shower_type, E / units.eV, iN))
            ax.set_xlabel('time [ns]')
            ax2.set_xlabel('time [ns]')
            ax.set_ylabel('normalized amplitude')
            ax2.semilogy(True)
            fig.tight_layout()
            fig2.tight_layout()
#                         plt.show()
            fig.savefig("plots/{}_{:.2g}eV_{:03d}.png".format(shower_type, E / units.eV, iN))
            fig2.savefig("plots/{}_{:.2g}eV_{:03d}_log.png".format(shower_type, E / units.eV, iN))
            plt.close("all")
    print("finished with  {} {:.3g}".format(shower_type, E))
    with open("ARZ_library_v1.1_{}_{:.1f}.pkl".format(shower_type, np.log10(E)), 'wb') as fout:
        pickle.dump(showers, fout, protocol=2)

if __name__ == "__main__":
    print('generating Askaryan pulses for the following viewing angles')
    print(dCs/units.deg)
    
    ps = []
    showers = {}
    lib_path = "/Users/cglaser/work/ARIANNA/data/ARZ/v1.1/library_v1.1.pkl"
    
    pool = Pool(processes=4)
    
    
    lib = io_utilities.read_pickle(lib_path)
    for iS, shower_type in enumerate(lib):  # loop through shower types
        if(shower_type not in showers):
            showers[shower_type] = {}
        print("shower type {}".format(shower_type))
        b = []
        nb = []
        for iE, E in enumerate(lib[shower_type]):  # loop through energies
            print('E = {:.2g}eV'.format(E))
            
#                 p = Process(target=calculate, args=(shower_type, E, len(lib[shower_type][E]['charge_excess'])))
#                 p.start()
#                 ps.append(p)
            r = pool.apply_async(calculate, args=(shower_type, E, len(lib[shower_type][E]['charge_excess'])))
            ps.append(r)
    for res in ps:
        print(res.get())
                
                
                                



import numpy as np
from NuRadioMC.utilities import units
from scipy import constants

e_mass = constants.physical_constants['electron mass energy equivalent in MeV'][0] * units.MeV
mu_mass = constants.physical_constants['muon mass energy equivalent in MeV'][0] * units.MeV
# Mass energy equivalent of the tau lepton
tau_mass = constants.physical_constants['tau mass energy equivalent in MeV'][0] * units.MeV
pi_mass = 139.57061 * units.MeV
rho770_mass = 775.49 * units.MeV
rho1450_mass = 1465 * units.MeV
a1_mass = 1230 * units.MeV
cspeed = constants.c * units.m / units.s
G_F = constants.physical_constants['Fermi coupling constant'][0] * units.GeV**(-2)

def get_neutrino_inelasticity(n_events):
    """
    Standard inelasticity for deep inelastic scattering used so far.
    Ported from ShelfMC

    Parameters:
    -----------
    n_events: int
        Number of events to be returned

    Returns:
    --------
    inelasticies: array
        Array with the inelasticities
    """
    R1 = 0.36787944
    R2 = 0.63212056
    inelasticities = (-np.log(R1 + np.random.uniform(0., 1., n_events) * R2)) ** 2.5

    return inelasticities

def get_ccnc(n_events):
    """
    Get the nature of the interaction current: cc or nc
    Ported from Shelf MC

    Parameters:
    -----------
    n_events: int
        Number of events to be returned

    Returns:
    --------
    ccnc: array
        Array with 'cc' or 'nc'
    """

    rnd = np.random.uniform(0., 1., n_events)
    ccnc = []
    for i, r in enumerate(rnd):
        #    if (r <= 0.6865254):#from AraSim
        if(r <= 0.7064):
            ccnc.append('cc')
        else:
            ccnc.append('nc')

    return np.array(ccnc)

def random_tau_branch():
    """
    Calculates a random tau branch decay
    See http://dx.doi.org/10.1016/j.cpc.2013.04.001

    Returns
    -------
    branch: string
        The corresponding decay branch
    """

    branching_ratios = np.array([0.18, 0.18])
    branching = np.random.uniform(0,1)
    if (branching < np.sum(branching_ratios[0:1])):
        # tau -> nu_tau + mu + nu_tau
        branch = 'tau_mu'
    elif (branching < np.sum(branching_ratios[0:2])):
        # tau -> nu_tau + e + nu_e
        branch = 'tau_e'
    else:
        # tau -> nu_tau + hadrons
        branch = 'tau_had'

    return branch


def inelasticity_tau_decay(tau_energy, branch):
    """
    Returns the hadronic or electromagnetic inelasticity for the tau decay
    See http://dx.doi.org/10.1016/j.cpc.2013.04.001
    and https://arxiv.org/pdf/1607.00193.pdf

    Parameters
    ----------
    tau_energy: float
        Tau energy at the moment of decay
    branch: string
        Type of tau decay: 'tau_mu', 'tau_e', 'tau_had'

    Returns
    -------
    inelasticity: float
        The fraction of energy carried by the leptonic or hadronic products
    """

    if ( branch == 'tau_had' ):

        branching = np.array([0.12, 0.26, 0.13, 0.13])
        rs = np.array([pi_mass, rho770_mass, a1_mass, rho1450_mass])/tau_mass

        def g_pi(y,r):
            if ( y < 0 or y > 1-r**2 ):
                return 0.
            else:
                return -(2*y-1+r)/(1-r**2)**2

        def g_1(y,r):
            if ( y < 0 or y > 1-r**2 ):
                return 0.
            else:
                return -(2*y-1+r)*(1-2*r)/(1-r)**2/(1+2*r)

        def g_0(y,r):
            if ( y < 0 or y > 1-r**2 ):
                return 0.
            else:
                return 1/(1-r)

        def y_distribution(y):
            pi_term = branching[0]*(g_pi(y,rs[0])+g_0(y,rs[0]))
            #rest_terms = branching[1:]*(g_1(y,rs)+g_0(y,rs))
            rest_terms = [ branch*(g_1(y,r)+g_0(y,r)) for branch,r in zip(branching[1:],rs[1:]) ]
            return pi_term + np.sum(rest_terms)

        chosen_y = rejection_sampling(y_distribution, 0, 1, 3)

        return 1-chosen_y

    elif ( branch == 'tau_e' or branch == 'tau_mu' ):

        mu = tau_mass
        if ( branch == 'tau_e' ):
            m_l = e_mass
        elif ( branch == 'tau_mu' ):
            m_l = mu_mass

        nu_min = m_l
        nu_max = (mu**2 + m_l**2)/2/mu

        # Fraction energy distibution in the decaying particle rest frame
        def x_distribution(x):
            if ( x < m_l/nu_max or x > 1 ):
                return 0.
            else:
                factor = G_F**2 * mu**5/192/np.pi**3
                return factor * (3-2*x)*x**2

        chosen_x = rejection_sampling(x_distribution, 0, 1, x_distribution(1))
        chosen_cos = np.random.uniform(-1,1)

        y_rest = chosen_x * nu_max/tau_mass
        # Transforming the rest inelasticity to the lab inelasticity
        y_lab = y_rest - np.sqrt(y_rest**2 - (m_l/mu)**2)*chosen_cos

        return y_lab

def rejection_sampling(f, xmin, xmax, ymax):
    """
    Draws a random number following a given distribution using
    a rejection sampling algorithm.

    Parameters
    ----------
    f: function
        Random distribution
    xmin: float
        Minimum value of the argument
    xmax: float
        Maximum value of the argument
    ymax: float
        Maximum function value to use for the rejection sample
        (e.g., the maximum of the function)

    Returns
    -------
    x: float
        Random value from the distribution
    """
    reject = True

    while( reject ):
        x = np.random.uniform(xmin, xmax)
        y = np.random.uniform(0, ymax)
        reject = f(x) < y

    return x



from __future__ import absolute_import, division, print_function
import numpy as np
import h5py
import argparse
import json

VERSION = 0.1

parser = argparse.ArgumentParser(description='Parse ARA event list.')
parser.add_argument('inputfilename', type=str,
                    help='path to (hdf5) input event list')
parser.add_argument('observerlist', type=str,
                    help='path to file containing the detector positions')
parser.add_argument('outputfilename', type=str,
                    help='name of output file storing the electric field traces at detector positions')
args = parser.parse_args()

# read in detector positions
with open(args.observerlist) as fobs:
    station_list = json.load(fobs)

fout = h5py.File(args.outputfilename, 'w')

fin = h5py.File(args.inputfilename, 'r')
for event in fin['eventlist']:  # loop through all events in eventlist
    evid, nuflavorint, nu_nubar, pnu, currentint, posnu_r, posnu_theta, posnu_phi, nnu_theta, nnu_phi, elast_y, = event

    evtout = fout.create_group('event_{:06d}'.format(evid))
    evtout.attrs['event_id'] = evid

    for station in station_list:
        # create dummy electric field pulse for each detector station
        n = 1024  # length of efield trace
        sampling = 1  # in ns
        efield = np.zeros((n, 4))
        efield[:, 0] = np.arange(0, sampling * n, sampling)  # the first collumn is the time array
        efield[400, 1] = 1  # add delta pulse to efield trace

        evtout[station['name']] = efield
        evtout[station['name']].attrs['position'] = station['position']  # for convenience save station position as attribute

fout.close()



from NuRadioMC.EvtGen.generator import *
import matplotlib.pyplot as plt
import json
plt.tight_layout()

"""
This file calculates some percentiles for the tau length and energy distributions
given by the CSDA, plots them, and saves them to a json file.

Just run:
    python T03_tau_length.py
"""

def get_p(p):

    return -tau_rest_lifetime * np.log(1-p)

def make_dict(keylist, valuelist):

    outdict = {}
    for key, value in zip(keylist, valuelist):
        outdict[key] = value

    return outdict


#user_times = np.array([])*units.fs
ps = [0.1, 0.5, 1-1/np.e, 0.9]
user_times = get_p(np.array(ps))

colourlist = ['orange', 'blue', 'red', 'black']
colours = make_dict(user_times, colourlist)

linestylelist = ['-','-','-.','-']
linestyles = make_dict(user_times, linestylelist)

labels = make_dict(user_times, [r'10%', 'Median', 'Mean', r'90%'])

energies = np.linspace(15.1, 20, 40)
energies = 10**energies
print(energies)

lengths = {}
lengths_nolosses = {}
tau_energies = {}

for user_time in user_times:
    lengths[user_time] = []
    lengths_nolosses[user_time] = []
    tau_energies[user_time] = []
    for energy in energies:
        print(energy)
        times = get_decay_time_losses(energy, 1000*units.km, average=True, compare=True, user_time=user_time)
        lengths[user_time].append(times[0]*cspeed)
        lengths_nolosses[user_time].append(times[1]*cspeed)
        tau_energies[user_time].append(times[2])

    lengths[user_time] = np.array(lengths[user_time])
    lengths_nolosses[user_time] = np.array(lengths_nolosses[user_time])


plt.loglog(energies, lengths[user_times[2]]/units.km, linestyle='-', color=colours[user_times[2]], label='Mean, with PN losses')
plt.loglog(energies, lengths_nolosses[user_times[2]]/units.km, linestyle='--', color=colours[user_times[2]], label='Mean, without losses')

plt.gcf().subplots_adjust(bottom=0.13)
fontsize=16
plt.tick_params(labelsize=12)
plt.fill_between(energies, lengths[user_times[0]]/units.km, lengths[user_times[-1]]/units.km, facecolor='0.75', interpolate=True, label=r'10% to 90% quantiles')
#plt.loglog([],[], linestyle='', label='P.N. losses - Solid\nNo losses - Dashed')
#plt.loglog([],[], linestyle='', label='No losses - Dashed')
plt.xlabel('Tau energy [eV]', size=fontsize)
plt.ylabel('Tau track length [km]', size=fontsize)
plt.legend(fontsize=12)
plt.savefig('tau_decay_length.png', format='png')
plt.show()

plt.gcf().subplots_adjust(bottom=0.13)
plt.tick_params(labelsize=12)
#for user_time in user_times:
#    plt.loglog(energies, tau_energies[user_time], linestyle='-', color=colours[user_time], label=labels[user_time])
#    if (user_time == user_times[2]):
#        plt.loglog(energies, tau_energies[user_time], linestyle=linestyles[user_time], color=colours[user_time], label='Mean, with PN losses')
plt.loglog(energies, tau_energies[user_times[2]], linestyle='-', color=colours[user_times[2]], label='Mean, with PN losses')
plt.loglog(energies, energies, linestyle='--', color=colours[user_times[2]], label="Mean, without losses")
plt.fill_between(energies, tau_energies[user_times[0]], tau_energies[user_times[-1]], facecolor='0.75', interpolate=True, label=r'10% to 90% quantiles')
#plt.loglog([],[], linestyle='', label='P.N. losses - Solid\nNo losses - Dashed')
#plt.loglog([],[], linestyle='', label='No losses - Dashed')
plt.xlabel('Tau initial energy [eV]', size=fontsize)
plt.ylabel('Tau decay energy [eV]', size=fontsize)
plt.legend(fontsize=12)
plt.savefig('tau_decay_energy.png', format='png')
plt.show()

output = {}
output['energies'] = list(energies)
output['lengths_10'] = list(lengths[user_times[0]])
output['lengths_90'] = list(lengths[user_times[-1]])
output['lengths_mean'] = list(lengths[user_times[2]])
output['tau_energies_10'] = list(tau_energies[user_times[0]])
output['tau_energies_90'] = list(tau_energies[user_times[-1]])
output['tau_energies_mean'] = list(tau_energies[user_times[2]])

with open('continuous_decay.json', 'w+') as data_file:
    json.dump(output, data_file, sort_keys=True, indent=4)



from __future__ import absolute_import, division, print_function, unicode_literals
import matplotlib.pyplot as plt
import logging
logging.basicConfig(level=logging.WARNING)
import numpy as np
from scipy import signal
import argparse
from datetime import datetime
from NuRadioReco.utilities import units, io_utilities
from radiotools import plthelpers as php
plt.switch_backend('agg')


i = 2

results = io_utilities.read_pickle("sim_results_{:02d}.pkl".format(i), encoding='latin1')
d = results['depth']
zen = np.array(results['exp'])[:, 0]
az = np.array(results['exp'])[:, 1]

fig, (ax, ax2) = plt.subplots(2, 1, sharex=True)
ax.scatter(d, (np.array(results['corr_LPDA'])[:, 0] - zen) / units.deg, label='corr LPDA',
           s=10, marker='o', alpha=0.5)
ax.scatter(d, (np.array(results['corr_dipole'])[:, 0] - zen) / units.deg, label='corr dipole',
           s=10, marker='d', alpha=0.5)
#     ax.scatter(d, (np.array(results['time_LPDA'])[:, 0] - zen) / units.deg, label='times LPDA',
#                s=10, marker='^', alpha=0.5)
#     ax.scatter(d, (np.array(results['time_dipole'])[:, 0] - zen) / units.deg, label='times dipole',
#                s=10, marker='x', alpha=0.5)
ax.set_ylabel("zenith (reco - exp) [deg]")
ax.legend()

ax2.scatter(d, (np.array(results['corr_LPDA'])[:, 1] - az) / units.deg, label='corr LPDA',
            s=10, marker='o', alpha=0.5)
ax2.scatter(d, (np.array(results['corr_dipole'])[:, 1] - az) / units.deg, label='corr dipole',
            s=10, marker='d', alpha=0.5)
#     ax2.scatter(d, (np.array(results['time_LPDA'])[:, 1] - az) / units.deg, label='times LPDA',
#                 s=10, marker='^', alpha=0.5)
#     ax2.scatter(d, (np.array(results['time_dipole'])[:, 1] - az) / units.deg, label='times dipole',
#                 s=10, marker='x', alpha=0.5)
ax2.set_ylabel("azimuth (reco - exp) [deg]")
ax2.set_xlabel("depth [m]")
fig.tight_layout()



fig, ax = php.get_histograms([(np.array(results['corr_LPDA'])[:, 0] - zen) / units.deg,
                    (np.array(results['corr_dipole'])[:, 0] - zen) / units.deg,
                    (np.array(results['time_LPDA'])[:, 0] - zen) / units.deg,
                    (np.array(results['time_dipole'])[:, 0] - zen) / units.deg,
                    (np.array(results['corr_LPDA'])[:, 1] - az) / units.deg,
                    (np.array(results['corr_dipole'])[:, 1] - az) / units.deg,
                    (np.array(results['time_LPDA'])[:, 1] - az) / units.deg,
                    (np.array(results['time_dipole'])[:, 1] - az) / units.deg],
                    bins=np.linspace(-15, 15, 30),
                    titles=['corr LPDA', 'corr_dipole', 'time_LPDA', 'time_dipole', '','','',''],
                    xlabels=[r'$\Delta$zen [deg]', r'$\Delta$zen [deg]', r'$\Delta$zen [deg]', r'$\Delta$zen [deg]',
                             r'$\Delta$az [deg]', r'$\Delta$az [deg]', r'$\Delta$az [deg]', r'$\Delta$az [deg]'])
fig.savefig("plots/results_{:02d}_hist.png".format(i))
plt.show()



from NuRadioMC.SignalGen.askaryan import get_time_trace, get_frequency_spectrum
from NuRadioMC.utilities import units, fft
import numpy as np
import pickle

np.random.seed(0)

E = 1e20 * units.eV
n_index = 1.78
domega = 0.05 * units.deg
theta = np.arccos(1. / n_index) + domega

dt = 0.5 * units.ns
n_samples = 256
ff = np.fft.rfftfreq(n_samples, dt)
tt = np.arange(0, n_samples * dt, dt)
R = 1 * units.km

models = ['Alvarez2009', 'ARZ2019', 'Alvarez2000']
shower_types = ['EM', 'HAD']

Es = 10**np.linspace(15,19,5) * units.eV
domegas = np.linspace(-5,5,10) * units.deg
thetas = np.arccos(1./n_index) + domegas

output = []
for model in models:
    for E in Es:
        for shower_type in shower_types:
            for theta in thetas:
                trace = get_time_trace(E, theta, n_samples, dt, shower_type, n_index, R, model)
                output.append(trace)

with open("reference_v1.pkl", "wb") as fout:
    pickle.dump(output, fout, protocol=2)


from __future__ import absolute_import, division, print_function
import argparse
# import detector simulation modules
import NuRadioReco.modules.efieldToVoltageConverter
import NuRadioReco.modules.ARIANNA.triggerSimulator
import NuRadioReco.modules.triggerSimulator
import NuRadioReco.modules.channelResampler
import NuRadioReco.modules.channelBandPassFilter
import NuRadioReco.modules.channelGenericNoiseAdder
from NuRadioReco.utilities import units
from NuRadioMC.simulation import as simulation
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("runstrawman")

# initialize detector sim modules
efieldToVoltageConverter = NuRadioReco.modules.efieldToVoltageConverter.efieldToVoltageConverter()
efieldToVoltageConverter.begin(debug=False, time_resolution=1 * units.ns,
                                         pre_pulse_time=0 * units.ns, post_pulse_time=0 * units.ns)
triggerSimulator = NuRadioReco.modules.triggerSimulator.triggerSimulator()
triggerSimulatorARIANNA = NuRadioReco.modules.ARIANNA.triggerSimulator.triggerSimulator()
channelResampler = NuRadioReco.modules.channelResampler.channelResampler()
channelBandPassFilter = NuRadioReco.modules.channelBandPassFilter.channelBandPassFilter()
channelGenericNoiseAdder = NuRadioReco.modules.channelGenericNoiseAdder.channelGenericNoiseAdder()


class mySimulation(simulation.simulation):

    def _detector_simulation(self):
        # start detector simulation
        efieldToVoltageConverter.run(self._evt, self._station, self._det)  # convolve efield with antenna pattern
        # downsample trace to internal simulation sampling rate (the efieldToVoltageConverter upsamples the trace to
        # 20 GHz by default to achive a good time resolution when the two signals from the two signal paths are added) 
        channelResampler.run(self._evt, self._station, self._det, sampling_rate=1. / self._dt)
        
        if bool(self._cfg['noise']):
            Vrms = self._Vrms / (self._bandwidth /( 2.5 * units.GHz))** 0.5  # normalize noise level to the bandwidth its generated for
            channelGenericNoiseAdder.run(self._evt, self._station, self._det, amplitude=Vrms, min_freq=0 * units.MHz,
                                         max_freq=2.5 * units.GHz, type='rayleigh')
        
        # bandpass filter trace, the upper bound is higher then the sampling rate which makes it just a highpass filter
        channelBandPassFilter.run(self._evt, self._station, self._det, passband=[80 * units.MHz, 500 * units.MHz],
                                  filter_type='butter', order=10)
        triggerSimulator.run(self._evt, self._station, self._det,
                             threshold=1 * self._Vrms,
                             triggered_channels=[0, 1, 2, 3],
                             number_concidences=1,
                             trigger_name='pre_trigger_1sigma')
        
        # downsample trace back to detector sampling rate
        channelResampler.run(self._evt, self._station, self._det, sampling_rate=self._sampling_rate_detector)


parser = argparse.ArgumentParser(description='Run NuRadioMC simulation')
parser.add_argument('inputfilename', type=str,
                    help='path to NuRadioMC input event list')
parser.add_argument('detectordescription', type=str,
                    help='path to file containing the detector description')
parser.add_argument('config', type=str,
                    help='NuRadioMC yaml config file')
parser.add_argument('outputfilename', type=str,
                    help='hdf5 output filename')
parser.add_argument('outputfilenameNuRadioReco', type=str, nargs='?', default=None,
                    help='outputfilename of NuRadioReco detector sim file')
args = parser.parse_args()

sim = mySimulation(eventlist=args.inputfilename,
                            outputfilename=args.outputfilename,
                            detectorfile=args.detectordescription,
                            config_file=args.config,
                            outputfilenameNuRadioReco=args.outputfilenameNuRadioReco)
sim.run()



from NuRadioMC.EvtGen.generator import *
import matplotlib.pyplot as plt

def get_p(p):

    return -tau_rest_lifetime * np.log(1-p)

table = create_interp('../decay_library.hdf5')

time = tau_rest_lifetime
times = np.linspace(0, 10*tau_rest_lifetime, 50)
energies = np.linspace(15,20,50)
energies = 10**energies

ps = [0.1, 0.5, 1-1/np.e, 0.9]
#ps = np.linspace(0.5,0.6,7)
times = get_p(np.array(ps))

exact = True
if exact:
    decay_times = []
    decay_energies = []
    for time in times:
        row_times = []
        row_energies = []
        for energy in energies:
            res = get_decay_time_losses(energy, 1000*units.km, average=False, compare=False, user_time=time)
            row_times.append(res[0])
            row_energies.append(res[1])
        decay_times.append(row_times)
        decay_energies.append(row_energies)

for itime, time in enumerate(times):
    plt.loglog(energies, table[0](time, energies)[0])
    if exact:
        plt.loglog(energies, decay_times[itime], linestyle='--')
plt.show()

for itime, time in enumerate(times):
    plt.loglog(energies, table[1](time, energies)[0])
    if exact:
        plt.loglog(energies, decay_energies[itime], linestyle='--')
plt.show()
